# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Operators and functions for linearized deformation."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import super

import numpy as np

from odl.discr import (DiscreteLp, DiscreteLpElement, Gradient, Divergence,
                       PointwiseInner)
from odl.operator.operator import Operator
from odl.space import ProductSpaceElement


__all__ = ('LinDeformFixedTempl', 'LinDeformFixedDisp')


def _linear_deform(template, displacement, out=None):
    """Linearized deformation of a template with a displacement field.

    The function maps a given template ``I`` and a given displacement
    field ``v`` to the new function ``x --> I(x + v(x))``.

    Parameters
    ----------
    template : `DiscreteLpElement`
        Template to be deformed by a displacement field.
    displacement : element of power space of ``template.space``
        Vector field (displacement field) used to deform the
        template.
    out : `numpy.ndarray`, optional
        Array to which the function values of the deformed template
        are written. It must have the same shape as ``template`` and
        a data type compatible with ``template.dtype``.

    Returns
    -------
    deformed_template : `numpy.ndarray`
        Function values of the deformed template. If ``out`` was given,
        the returned object is a reference to it.

    Examples
    --------
    Create a simple 1D template to initialize the operator and
    apply it to a displacement field. Where the displacement is zero,
    the output value is the same as the input value.
    In the 4-th point, the value is taken from 0.2 (one cell) to the
    left, i.e. 1.0.

    >>> space = odl.uniform_discr(0, 1, 5)
    >>> disp_field_space = space.vector_field_space
    >>> template = space.element([0, 0, 1, 0, 0])
    >>> displacement_field = disp_field_space.element([[0, 0, 0, -0.2, 0]])
    >>> _linear_deform(template, displacement_field)
    array([ 0.,  0.,  1.,  1.,  0.])

    The result depends on the chosen interpolation. With 'linear'
    interpolation and an offset equal to half the distance between two
    points, 0.1, one gets the mean of the values.

    >>> space = odl.uniform_discr(0, 1, 5, interp='linear')
    >>> disp_field_space = space.vector_field_space
    >>> template = space.element([0, 0, 1, 0, 0])
    >>> displacement_field = disp_field_space.element([[0, 0, 0, -0.1, 0]])
    >>> _linear_deform(template, displacement_field)
    array([ 0. ,  0. ,  1. ,  0.5,  0. ])
    """
    image_pts = template.space.points()
    for i, vi in enumerate(displacement):
        image_pts[:, i] += vi.ntuple.asarray()
    return template.interpolation(image_pts.T, out=out, bounds_check=False)


class LinDeformFixedTempl(Operator):

    """Deformation operator with fixed template acting on displacement fields.

    The operator has a fixed template ``I`` and maps a displacement
    field ``v`` to the new function ``x --> I(x + v(x))``.

    See Also
    --------
    LinDeformFixedDisp : Deformation with a fixed displacement.

    Notes
    -----
    For :math:`\Omega \subset \mathbb{R}^d`, we take :math:`X = L^p(\Omega)`
    to be the template space, i.e. :math:`I \\in X`. Then the vector field
    space is identified with :math:`V := X^d`. Hence the deformation operator
    with fixed template maps :math:`V` into :math:`X`:

    .. math::

        W_I : V \\to X, \quad W_I(v) := I(\cdot + v(\cdot)),

    i.e., :math:`W_I(v)(x) = I(x + v(x))`.

    Note that this operator is non-linear. Its derivative at :math:`v` is
    an operator that maps :math:`V` into :math:`X`:

    .. math::

        W_I'(v) : V \\to X, \quad W_I'(v)(u) =
        \\big< \\nabla I(\cdot + v(\cdot)), u \\big>_{\mathbb{R}^d},

    i.e., :math:`W_I'(v)(u)(x) = \\nabla I(x + v(x))^T u(x)`,

    which is to be understood as a point-wise inner product, resulting
    in a function in :math:`X`. And the adjoint of the preceding derivative
    is also an operator that maps :math:`X` into :math:`V`:

    .. math::

        W_I'(v)^* : X \\to V, \quad W_I'(v)^*(J) =
        J \, \\nabla I(\cdot + v(\cdot)),

    i.e., :math:`W_I'(v)^*(J)(x) = J(x) \, \\nabla I(x + v(x))`.
    """

    def __init__(self, template):
        """Initialize a new instance.

        Parameters
        ----------
        template : `DiscreteLpElement`
            Fixed template that is to be deformed.

        Examples
        --------
        Create a simple 1D template to initialize the operator and
        apply it to a displacement field. Where the displacement is zero,
        the output value is the same as the input value.
        In the 4-th point, the value is taken from 0.2 (one cell) to the
        left, i.e. 1.0.

        >>> space = odl.uniform_discr(0, 1, 5, interp='nearest')
        >>> template = space.element([0, 0, 1, 0, 0])
        >>> op = LinDeformFixedTempl(template)
        >>> disp_field = [[0, 0, 0, -0.2, 0]]
        >>> print(op(disp_field))
        [0.0, 0.0, 1.0, 1.0, 0.0]

        The result depends on the chosen interpolation. With 'linear'
        interpolation and an offset equal to half the distance between two
        points, 0.1, one gets the mean of the values.

        >>> space = odl.uniform_discr(0, 1, 5, interp='linear')
        >>> template = space.element([0, 0, 1, 0, 0])
        >>> op = LinDeformFixedTempl(template)
        >>> disp_field = [[0, 0, 0, -0.1, 0]]
        >>> print(op(disp_field))
        [0.0, 0.0, 1.0, 0.5, 0.0]
        """
        if not isinstance(template, DiscreteLpElement):
            raise TypeError('`template` must be a `DiscreteLpElement`'
                            'instance, got {!r}'.format(template))

        self.__template = template
        super().__init__(domain=self.template.space.vector_field_space,
                         range=self.template.space,
                         linear=False)

    @property
    def template(self):
        """Fixed template of this deformation operator."""
        return self.__template

    def _call(self, displacement, out=None):
        """Implementation of ``self(displacement[, out])``."""
        return _linear_deform(self.template, displacement, out)

    def derivative(self, displacement):
        """Derivative of the operator at ``displacement``.

        Parameters
        ----------
        displacement : `domain` `element-like`
            Point at which the derivative is computed.

        Returns
        -------
        derivative : `PointwiseInner`
            The derivative evaluated at ``displacement``.
        """
        # To implement the complex case we need to be able to embed the real
        # vector field space into the range of the gradient. Issue #59.
        if not self.range.is_rn:
            raise NotImplementedError('derivative not implemented for complex '
                                      'spaces.')

        displacement = self.domain.element(displacement)

        # TODO: allow users to select what method to use here.
        grad = Gradient(domain=self.range, method='central',
                        pad_mode='symmetric')
        grad_templ = grad(self.template)
        def_grad = self.domain.element(
            [_linear_deform(gf, displacement) for gf in grad_templ])

        return PointwiseInner(self.domain, def_grad)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r})'.format(self.__class__.__name__, self.template)


class LinDeformFixedDisp(Operator):

    """Deformation operator with fixed displacement acting on templates.

    The operator has a fixed displacement field ``v`` and maps a template
    ``I`` to the new function ``x --> I(x + v(x))``.

    See Also
    --------
    LinDeformFixedTempl : Deformation with a fixed template.

    Notes
    -----
    For :math:`\Omega \subset \mathbb{R}^d`, we take :math:`V := X^d`
    to be the space of displacement fields, where :math:`X = L^p(\Omega)`
    is the template space. Hence the deformation operator with the fixed
    displacement field :math:`v \\in V` maps :math:`X` into :math:`X`:

    .. math::

        W_v : X \\to X, \quad W_v(I) := I(\cdot + v(\cdot)),

    i.e., :math:`W_v(I)(x) = I(x + v(x))`.

    This operator is linear, so its derivative is itself, but it may not be
    bounded and may thus not have a formal adjoint. For "small" :math:`v`,
    though, one can approximate the adjoint by

    .. math::

        W_v^*(I) \\approx \exp(-\mathrm{div}\, v) \, I(\cdot - v(\cdot)),

    i.e., :math:`W_v^*(I)(x) \\approx \exp(-\mathrm{div}\,v(x))\, I(x - v(x))`.
    """

    def __init__(self, displacement, templ_space=None):
        """Initialize a new instance.

        Parameters
        ----------
        displacement : element of a power space of `DiscreteLp`
            Fixed displacement field used in the deformation.
        templ_space : `DiscreteLp`, optional
            Template space on which this operator is applied, i.e. the
            operator domain and range. It must fulfill
            ``domain.vector_field_space == displacement.space``, so this
            option is useful mainly for support of complex spaces.
            Default: ``displacement.space[0]``

        Examples
        --------
        Create a simple 1D template to initialize the operator and
        apply it to a displacement field. Where the displacement is zero,
        the output value is the same as the input value.
        In the 4-th point, the value is taken from 0.2 (one cell) to the
        left, i.e. 1.0.

        >>> space = odl.uniform_discr(0, 1, 5)
        >>> disp_field = space.vector_field_space.element([[0, 0, 0, -0.2, 0]])
        >>> op = LinDeformFixedDisp(disp_field)
        >>> template = [0, 0, 1, 0, 0]
        >>> print(op([0, 0, 1, 0, 0]))
        [0.0, 0.0, 1.0, 1.0, 0.0]

        The result depends on the chosen interpolation. With 'linear'
        interpolation and an offset equal to half the distance between two
        points, 0.1, one gets the mean of the values.

        >>> space = odl.uniform_discr(0, 1, 5, interp='linear')
        >>> disp_field = space.vector_field_space.element([[0, 0, 0, -0.1, 0]])
        >>> op = LinDeformFixedDisp(disp_field)
        >>> template = [0, 0, 1, 0, 0]
        >>> print(op(template))
        [0.0, 0.0, 1.0, 0.5, 0.0]
        """
        if not isinstance(displacement, ProductSpaceElement):
            raise TypeError('`displacement must be a `ProductSpaceElement` '
                            'instance, got {!r}'.format(displacement))
        if not displacement.space.is_power_space:
            raise TypeError('`displacement.space` must be a power space, '
                            'got {!r}'.format(displacement.space))
        if not isinstance(displacement.space[0], DiscreteLp):
            raise TypeError('`displacement.space[0]` must be a `DiscreteLp` '
                            'instance, got {!r}'.format(displacement.space[0]))

        if templ_space is None:
            templ_space = displacement.space[0]
        else:
            if not isinstance(templ_space, DiscreteLp):
                raise TypeError('`templ_space` must be a `DiscreteLp` '
                                'instance, got {!r}'.format(templ_space))
            if templ_space.vector_field_space != displacement.space:
                raise ValueError('`templ_space.vector_field_space` not equal '
                                 'to `displacement.space` ({} != {})'
                                 ''.format(templ_space.vector_field_space,
                                           displacement.space))

        self.__displacement = displacement
        super().__init__(domain=templ_space, range=templ_space, linear=True)

    @property
    def displacement(self):
        """Fixed displacement field of this deformation operator."""
        return self.__displacement

    def _call(self, template, out=None):
        """Implementation of ``self(template[, out])``."""
        return _linear_deform(template, self.displacement, out)

    @property
    def inverse(self):
        """Inverse deformation using ``-v`` as displacement.

        Note that this implementation uses an approximation that is only
        valid for small displacements.
        """
        return LinDeformFixedDisp(-self.displacement, templ_space=self.domain)

    @property
    def adjoint(self):
        """Adjoint of the linear operator.

        Note that this implementation uses an approximation that is only
        valid for small displacements.
        """
        # TODO allow users to select what method to use here.
        div_op = Divergence(domain=self.displacement.space, method='forward',
                            pad_mode='symmetric')
        jacobian_det = self.domain.element(
            np.exp(-div_op(self.displacement)))

        return jacobian_det * self.inverse

    def __repr__(self):
        """Return ``repr(self)``."""
        if self.domain == self.__displacement.space[0]:
            domain_repr = ''
        else:
            domain_repr = ', domain={!r}'.format(self.domain)

        return '{}({!r}{})'.format(self.__class__.__name__,
                                   self.__displacement,
                                   domain_repr)

if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()

import pytest


__all__ = ('skip_if_no_astra', 'skip_if_no_astra_cuda', 'skip_if_no_scikit')


skip_if_no_astra = pytest.mark.skipif("not odl.tomo.ASTRA_AVAILABLE",
                                      reason='ASTRA not available')

skip_if_no_astra_cuda = pytest.mark.skipif("not odl.tomo.ASTRA_CUDA_AVAILABLE",
                                           reason='ASTRA CUDA not available')

skip_if_no_scikit = pytest.mark.skipif("not odl.tomo.SCIKIT_IMAGE_AVAILABLE",
                                       reason='scikit not available')



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Setup script for ODL.

Installation command::

    pip install [--user] [-e] .
"""

from __future__ import print_function, absolute_import

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import os
import sys


if os.environ.get('READTHEDOCS', None) == 'True':
    # Mock requires in conf.py
    requires = ''
    test_requires = []
else:
    requires = open(
        os.path.join(os.path.dirname(__file__),
                     'requirements.txt')).readlines()
    test_requires = open(
        os.path.join(os.path.dirname(__file__),
                     'test_requirements.txt')).readlines()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

long_description = """
Operator Discretization Library (ODL) is a Python library for fast prototyping focusing on (but not restricted to) inverse problems. ODL is being developed at `KTH Royal Institute of Technology <https://www.kth.se/en/sci/institutioner/math>`_.

The main intent of ODL is to enable mathematicians and applied scientists to use different numerical methods on real-world problems without having to implement all necessary parts from the bottom up.
This is reached by an `Operator` structure which encapsulates all application-specific parts, and a high-level formulation of solvers which usually expect an operator, data and additional parameters.
The main advantages of this approach is that

1. Different problems can be solved with the same method (e.g. TV regularization) by simply switching operator and data.
2. The same problem can be solved with different methods by simply calling into different solvers.
3. Solvers and application-specific code need to be written only once, in one place, and can be tested individually.
4. Adding new applications or solution methods becomes a much easier task.



Features
========

- Efficient and well-tested data containers based on Numpy (default) or CUDA (optional)
- Objects to represent mathematical notions like vector spaces and operators, including properties as expected from mathematics (inner product, norm, operator composition, ...)
- Convenience functionality for operators like arithmetic, composition, operator matrices etc., which satisfy the known mathematical rules.
- Out-of-the-box support for frequently used operators like scaling, partial derivative, gradient, Fourier transform etc.
- Support for tomographic imaging with a unified geometry representation and bindings to external libraries for efficient computation of projections and back-projections.
- Standardized tests to validate implementations against expected behavior of the corresponding mathematical object, e.g. if a user-defined norm satisfies `norm(x + y) <= norm(x) + norm(y)` for a number of input vectors `x` and `y`.
"""

setup(
    name='odl',

    version='0.5.0',

    description='Operator Discretization Library',
    long_description=long_description,

    url='https://github.com/odlgroup/odl',

    author='ODL development group',
    author_email='odl@math.kth.se',

    license='GPLv3+',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development',

        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',

        'Operating System :: OS Independent'
    ],

    keywords='research development mathematics prototyping imaging tomography',

    packages=find_packages(exclude=['*test*']),
    package_dir={'odl': 'odl'},

    install_requires=[requires],
    tests_require=['pytest'],
    extras_require={
        'testing': test_requires,
        'show': 'matplotlib',
        'fftw': 'pyfftw',
        'pywavelets': 'Pywavelets>=0.4',
        'scikit': 'scikit-image',
        'proximal': 'proximal',
    },

    cmdclass={'test': PyTest},
)



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""First-order primal-dual algorithm developed by Chambolle and Pock.

The Chambolle-Pock algorithm is a flexible method well suited for
non-smooth convex optimization problems in imaging. It was first
proposed in [CP2011a]_.
"""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()

import numpy as np

from odl.operator import Operator


__all__ = ('chambolle_pock_solver',)


# TODO: add dual gap as convergence measure
# TODO: diagonal preconditioning

def chambolle_pock_solver(x, f, g, L, tau, sigma, niter=1, **kwargs):
    """Chambolle-Pock algorithm for non-smooth convex optimization problems.

    First order primal-dual hybrid-gradient method for non-smooth convex
    optimization problems with known saddle-point structure. The
    primal formulation of the general problem is::

        min_{x in X} f(L x) + g(x)

    where ``L`` is an operator and ``F`` and ``G`` are functionals.

    The Chambolle-Pock algorithm is a primal-dual algorithm, and basically
    consists of alternating a gradient ascent in the dual variable and a
    gradient descent in the primal variable. The proximal operator is used to
    generate a ascent direction for the convex conjugate of F and descent
    direction for G. Additionally an over-relaxation in the primal variable is
    performed.

    Parameters
    ----------
    x : ``op.domain`` element
        Starting point of the iteration, updated in-place.
    f : `Functional`
        The function ``f`` in the problem definition. Needs to have
        ``f.convex_conj.proximal``.
    g : `Functional`
        The function ``f`` in the problem definition. Needs to have
        ``g.proximal``.
    L : linear `Operator`
        The linear operator that should be applied before ``f``. Its range must
        match the domain of ``f``.
    tau : positive float
        Step size parameter for the update of the primal (``g``) variable.
        Controls the speed of convergence towards the minimum of G.
    sigma : positive float
        Step size parameter for the update of the dual (``f``) variable.
        Controls the speed of convergence towards the minimum of ``F^*``.
    niter : non-negative int, optional
        Number of iterations.

    Other Parameters
    ----------------
    callback : callable, optional
        Function called with the current iterate after each iteration.
    theta : float, optional
        Relaxation parameter, required to fulfill ``0 <= theta <= 1``.
        Default: 1
    gamma : non-negative float, optional
        Acceleration parameter. If not ``None``, it overrides ``theta`` and
        causes variable relaxation parameter and step sizes to be used,
        with ``tau`` and ``sigma`` as initial values. Requires ``G`` or
        ``F^*`` to be uniformly convex.
        Default: ``None``
    x_relax : ``op.domain`` element, optional
        Required to resume iteration. For ``None``, a copy of the primal
        variable ``x`` is used.
        Default: ``None``
    y : ``op.range`` element, optional
        Required to resume iteration. For ``None``, ``op.range.zero()``
        is used.
        Default: ``None``

    Notes
    -----
    The problem of interest is

    .. math::

        \\min_{x \\in X} F(K x) + G(x),

    where the technical conditions are that :math:`K` is an operator
    between Hilbert spaces :math:`X` and :math:`Y`, where convergence is only
    guaranteed if :math:`K` is linear and :math:`X, Y` are finite dimensional.
    Further, :math:`G : X -> [0, +\\infty]` and :math:`F : Y -> [0, +\\infty]`
    are proper, convex, lower-semicontinuous functionals.

    It is often of interest to study problems that involve several operators,
    for example the classical TV regularized problem

    .. math::

        \\min_x ||Ax - b||_2^2 + ||\\nabla x||_1.

    Here it is tempting to let :math:`K=A`, :math:`F(y)=||y||_2^2` and
    :math:`G(x)=||\\nabla x||_1`. This is however not feasible since the
    proximal of :math:`||\\nabla x||_1` has no closed form expression.

    Instead, the problem can be formulated :math:`K(x) = (A(x), \\nabla x)`,
    :math:`F((x_1, x_2)) = ||x_1||_2^2 + ||x_2||_1`, :math:`G(x)=0`. See the
    examples folder for more information on how to do this.

    See Also
    --------
    odl.solvers.nonsmooth.douglas_rachford.douglas_rachford_pd :
        Solver for similar problems which can additionaly handle infimal
        convolutions and multiple forward operators.
    odl.solvers.nonsmooth.forward_backward.forward_backward_pd :
        Solver for similar problems which can additionaly handle infimal
        convolutions, multiple forward operators and a differentiable term.

    References
    ----------
    For a more detailed documentation see :ref:`chambolle_pock`.

    References on the Chambolle-Pock algorithm can be found in [CP2011a]_ and
    [CP2011b]_.

    This implementation of the CP algorithm is along the lines of
    [Sid+2012]_.

    For more on convex analysis including convex conjugates and
    resolvent operators see [Roc1970]_.

    For more on proximal operators and algorithms see [PB2014]_.

    The non-linear case is analyzed in [Val2014]_.
    """
    # Forward operator
    if not isinstance(L, Operator):
        raise TypeError('`op` {!r} is not an `Operator` instance'
                        ''.format(L))

    # Starting point
    if x not in L.domain:
        raise TypeError('`x` {!r} is not in the domain of `op` {!r}'
                        ''.format(x, L.domain))

    # Step size parameter
    tau, tau_in = float(tau), tau
    if tau <= 0:
        raise ValueError('`tau` must be positive, got {}'.format(tau_in))

    # Step size parameter
    sigma, sigma_in = float(sigma), sigma
    if sigma <= 0:
        raise ValueError('`sigma` must be positive, got {}'.format(sigma_in))

    # Number of iterations
    if not isinstance(niter, int) or niter < 0:
        raise ValueError('`niter` {} not understood'
                         ''.format(niter))

    # Relaxation parameter
    theta = kwargs.pop('theta', 1)
    theta, theta_in = float(theta), theta
    if not 0 <= theta <= 1:
        raise ValueError('`theta` {} not in [0, 1]'
                         ''.format(theta_in))

    # Acceleration parameter
    gamma = kwargs.pop('gamma', None)
    if gamma is not None:
        gamma, gamma_in = float(gamma), gamma
        if gamma < 0:
            raise ValueError('`gamma` must be non-negative, got {}'
                             ''.format(gamma_in))

    # Callback object
    callback = kwargs.pop('callback', None)
    if callback is not None and not callable(callback):
        raise TypeError('`callback` {} is not callable'
                        ''.format(callback))

    # Initialize the relaxation variable
    x_relax = kwargs.pop('x_relax', None)
    if x_relax is None:
        x_relax = x.copy()
    elif x_relax not in L.domain:
        raise TypeError('`x_relax` {} is not in the domain of '
                        '`L` {}'.format(x_relax.space, L.domain))

    # Initialize the dual variable
    y = kwargs.pop('y', None)
    if y is None:
        y = L.range.zero()
    elif y not in L.range:
        raise TypeError('`y` {} is not in the range of `L` '
                        '{}'.format(y.space, L.range))

    # Get the proximals
    proximal_dual = f.convex_conj.proximal
    proximal_primal = g.proximal

    # Temporary copy to store previous iterate
    x_old = x.space.element()

    for _ in range(niter):
        # Copy required for relaxation
        x_old.assign(x)

        # Gradient ascent in the dual variable y
        dual_tmp = y + sigma * L(x_relax)
        proximal_dual(sigma)(dual_tmp, out=y)

        # Gradient descent in the primal variable x
        primal_tmp = x + (- tau) * L.derivative(x).adjoint(y)
        proximal_primal(tau)(primal_tmp, out=x)

        # Acceleration
        if gamma is not None:
            theta = float(1 / np.sqrt(1 + 2 * gamma * tau))
            tau *= theta
            sigma /= theta

        # Over-relaxation in the primal variable x
        x_relax.lincomb(1 + theta, x, -theta, x_old)

        if callback is not None:
            callback(x)


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Backend for ASTRA using CUDA."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()

import numpy as np
try:
    import astra
    ASTRA_CUDA_AVAILABLE = astra.astra.use_cuda()
except ImportError:
    ASTRA_CUDA_AVAILABLE = False

from odl.discr import DiscreteLp, DiscreteLpElement
from odl.tomo.backends.astra_setup import (
    astra_projection_geometry, astra_volume_geometry, astra_projector,
    astra_data, astra_algorithm)
from odl.tomo.geometry import (
    Geometry, Parallel2dGeometry, FanFlatGeometry, Parallel3dAxisGeometry,
    HelicalConeFlatGeometry)
from odl.util.utility import writable_array


__all__ = ('astra_cuda_forward_projector', 'astra_cuda_back_projector',
           'ASTRA_CUDA_AVAILABLE')


# TODO: use context manager when creating data structures
# TODO: is magnification scaling at the right place?

def astra_cuda_forward_projector(vol_data, geometry, proj_space, out=None):
    """Run an ASTRA forward projection on the given data using the GPU.

    Parameters
    ----------
    vol_data : `DiscreteLpElement`
        Volume data to which the projector is applied
    geometry : `Geometry`
        Geometry defining the tomographic setup
    proj_space : `DiscreteLp`
        Space to which the calling operator maps
    out : ``proj_space`` element, optional
        Element of the projection space to which the result is written. If
        ``None``, an element in ``proj_space`` is created.

    Returns
    -------
    out : ``proj_space`` element
        Projection data resulting from the application of the projector.
        If ``out`` was provided, the returned object is a reference to it.
    """
    if not isinstance(vol_data, DiscreteLpElement):
        raise TypeError('volume data {!r} is not a DiscreteLpElement '
                        'instance'.format(vol_data))
    if not isinstance(geometry, Geometry):
        raise TypeError('geometry  {!r} is not a Geometry instance'
                        ''.format(geometry))
    if vol_data.ndim != geometry.ndim:
        raise ValueError('dimensions {} of volume data and {} of geometry '
                         'do not match'
                         ''.format(vol_data.ndim, geometry.ndim))
    if not isinstance(proj_space, DiscreteLp):
        raise TypeError('projection space {!r} is not a DiscreteLp '
                        'instance'.format(proj_space))
    if out is not None:
        if not isinstance(out, DiscreteLpElement):
            raise TypeError('`out` {} is neither None nor a '
                            'DiscreteLpElement instance'.format(out))

    ndim = vol_data.ndim

    # Create astra geometries
    vol_geom = astra_volume_geometry(vol_data.space)
    proj_geom = astra_projection_geometry(geometry)

    # Create ASTRA data structures

    # In the case dim == 3, we need to swap axes, so can't perform the FP
    # in-place
    vol_id = astra_data(vol_geom, datatype='volume', data=vol_data,
                        allow_copy=True)

    # Create projector
    proj_id = astra_projector('nearest', vol_geom, proj_geom, ndim,
                              impl='cuda')

    if ndim == 2:
        if out is None:
            out = proj_space.element()

        # Wrap the array in correct dtype etc if needed
        with writable_array(out, dtype='float32', order='C') as arr:
            sino_id = astra_data(proj_geom, datatype='projection', data=arr)

            # Create algorithm
            algo_id = astra_algorithm('forward', ndim, vol_id, sino_id,
                                      proj_id=proj_id, impl='cuda')

            # Run algorithm
            astra.algorithm.run(algo_id)
    elif ndim == 3:
        sino_id = astra_data(proj_geom, datatype='projection',
                             ndim=proj_space.ndim)

        # Create algorithm
        algo_id = astra_algorithm('forward', ndim, vol_id, sino_id,
                                  proj_id=proj_id, impl='cuda')

        # Run algorithm
        astra.algorithm.run(algo_id)

        if out is None:
            out = proj_space.element(np.rollaxis(astra.data3d.get(sino_id),
                                                 0, 3))
        else:
            out[:] = np.rollaxis(astra.data3d.get(sino_id), 0, 3)
    else:
        raise RuntimeError('unknown ndim')

    # Fix inconsistent scaling
    if isinstance(geometry, Parallel2dGeometry):
        # parallel2d scales with pixel stride
        out *= 1 / float(geometry.det_partition.cell_sides[0])

    # Delete ASTRA objects
    astra.algorithm.delete(algo_id)
    if ndim == 2:
        astra.data2d.delete((vol_id, sino_id))
        astra.projector.delete(proj_id)
    else:
        astra.data3d.delete((vol_id, sino_id))
        astra.projector3d.delete(proj_id)

    return out


def astra_cuda_back_projector(proj_data, geometry, reco_space, out=None):
    """Run an ASTRA backward projection on the given data using the GPU.

    Parameters
    ----------
    proj_data : `DiscreteLp` element
        Projection data to which the backward projector is applied
    geometry : `Geometry`
        Geometry defining the tomographic setup
    reco_space : `DiscreteLp`
        Space to which the calling operator maps
    out : ``reco_space`` element, optional
        Element of the reconstruction space to which the result is written.
        If ``None``, an element in ``reco_space`` is created.

    Returns
    -------
    out : ``reco_space`` element
        Reconstruction data resulting from the application of the backward
        projector. If ``out`` was provided, the returned object is a
        reference to it.
        """
    if not isinstance(proj_data, DiscreteLpElement):
        raise TypeError('projection data {!r} is not a DiscreteLpElement '
                        'instance'.format(proj_data))
    if not isinstance(geometry, Geometry):
        raise TypeError('geometry  {!r} is not a Geometry instance'
                        ''.format(geometry))
    if not isinstance(reco_space, DiscreteLp):
        raise TypeError('reconstruction space {!r} is not a DiscreteLp '
                        'instance'.format(reco_space))
    if reco_space.ndim != geometry.ndim:
        raise ValueError('dimensions {} of reconstruction space and {} of '
                         'geometry do not match'.format(reco_space.ndim,
                                                        geometry.ndim))
    if out is not None:
        if not isinstance(out, DiscreteLpElement):
            raise TypeError('`out` {} is neither None nor a '
                            'DiscreteLpElement instance'.format(out))

    ndim = proj_data.ndim

    # Create geometries
    vol_geom = astra_volume_geometry(reco_space)
    proj_geom = astra_projection_geometry(geometry)

    if ndim == 2:
        swapped_proj_data = proj_data
    else:
        swapped_proj_data = np.ascontiguousarray(
            np.rollaxis(proj_data.asarray(), 2, 0))

    sino_id = astra_data(proj_geom, datatype='projection',
                         data=swapped_proj_data, allow_copy=True)

    # Create projector
    proj_id = astra_projector('nearest', vol_geom, proj_geom, ndim,
                              impl='cuda')

    # Reconstruction volume
    if out is None:
        out = reco_space.element()

    # Wrap the array in correct dtype etc if needed
    with writable_array(out, dtype='float32', order='C') as arr:
        vol_id = astra_data(vol_geom, datatype='volume', data=arr,
                            ndim=reco_space.ndim)
        # Create algorithm
        algo_id = astra_algorithm('backward', ndim, vol_id, sino_id,
                                  proj_id=proj_id, impl='cuda')

        # Run algorithm
        astra.algorithm.run(algo_id)

    # Angular integration weighting factor
    # angle interval weight by approximate cell volume
    extent = float(geometry.motion_partition.extent())
    size = float(geometry.motion_partition.size)
    scaling_factor = extent / size

    # Fix inconsistent scaling
    if isinstance(geometry, Parallel2dGeometry):
        # scales with 1 / (cell volume)
        scaling_factor *= float(reco_space.cell_volume)
    elif isinstance(geometry, FanFlatGeometry):
        # scales with 1 / (cell volume)
        scaling_factor *= float(reco_space.cell_volume)
        # magnification correction
        src_radius = geometry.src_radius
        det_radius = geometry.det_radius
        scaling_factor *= ((src_radius + det_radius) / src_radius)
    elif isinstance(geometry, Parallel3dAxisGeometry):
        # scales with voxel stride
        # currently only square voxels are supported
        extent = reco_space.partition.extent()
        shape = np.array(reco_space.shape, dtype=float)
        scaling_factor /= float(extent[0] / shape[0])
    elif isinstance(geometry, HelicalConeFlatGeometry):
        # cone scales with voxel stride
        # currently only square voxels are supported
        extent = reco_space.partition.extent()
        shape = np.array(reco_space.shape, dtype=float)
        scaling_factor /= float(extent[0] / shape[0])
        # magnification correction
        src_radius = geometry.src_radius
        det_radius = geometry.det_radius
        scaling_factor *= ((src_radius + det_radius) / src_radius) ** 2
    out *= scaling_factor

    # Delete ASTRA objects
    astra.algorithm.delete(algo_id)
    if ndim == 2:
        astra.data2d.delete((vol_id, sino_id))
        astra.projector.delete(proj_id)
    else:
        astra.data3d.delete((vol_id, sino_id))
        astra.projector3d.delete(proj_id)

    return out


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Ray transforms."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import str, super

import numpy as np

from odl.discr import DiscreteLp
from odl.operator import Operator
from odl.space import FunctionSpace
from odl.tomo.geometry import Geometry, Parallel2dGeometry
from odl.tomo.backends import (
    ASTRA_AVAILABLE, ASTRA_CUDA_AVAILABLE,
    astra_cpu_forward_projector, astra_cpu_back_projector,
    astra_cuda_forward_projector, astra_cuda_back_projector,
    scikit_radon_forward, scikit_radon_back_projector)

_SUPPORTED_IMPL = ('astra_cpu', 'astra_cuda', 'scikit')


__all__ = ('RayTransform', 'RayBackProjection')


# TODO: DivergentBeamTransform?

class RayTransform(Operator):

    """Discrete Ray transform between L^p spaces."""

    def __init__(self, discr_domain, geometry, impl='astra_cpu', **kwargs):
        """Initialize a new instance.

        Parameters
        ----------
        discr_domain : `DiscreteLp`
            Discretized space, the domain of the forward projector
        geometry : `Geometry`
            Geometry of the transform, containing information about
            the operator range
        impl : {'astra_cpu', 'astra_cuda', 'scikit'}, optional
            Implementation back-end for the transform. Supported back-ends:
            'astra_cpu': ASTRA toolbox using CPU, only 2D
            'astra_cuda': ASTRA toolbox, using CUDA, 2D or 3D
            'scikit': scikit-image, only 2D parallel with square domain
        interp : {'nearest', 'linear'}
            Interpolation type for the discretization of the operator
            range.
            Default: 'nearest'

        Other Parameters
        ----------------
        discr_range : `DiscreteLp`
            Discretized space, the range of the forward projector.
            Default: Infered from parameters.

        Notes
        -----
        The ASTRA backend is faster if data is given with ``dtype`` 'float32'
        and storage order 'C'. Otherwise copies will be needed.
        """
        if not isinstance(discr_domain, DiscreteLp):
            raise TypeError('`discr_domain` {!r} is not a `DiscreteLp`'
                            ' instance'.format(discr_domain))

        if not isinstance(geometry, Geometry):
            raise TypeError('`geometry` {!r} is not a `Geometry` instance'
                            ''.format(geometry))

        impl, impl_in = str(impl).lower(), impl
        if impl not in _SUPPORTED_IMPL:
            raise ValueError('`impl` {!r} not supported'
                             ''.format(impl_in))

        # TODO: sanity checks between impl and discretization impl
        if impl.startswith('astra'):
            # TODO: these should be moved somewhere else
            if not ASTRA_AVAILABLE:
                raise ValueError("'astra' back-end not available")
            if impl == 'astra_cuda' and not ASTRA_CUDA_AVAILABLE:
                raise ValueError("'astra_cuda' back-end not available")
            if not np.allclose(discr_domain.partition.cell_sides[1:],
                               discr_domain.partition.cell_sides[:-1]):
                raise ValueError('ASTRA does not support different voxel '
                                 'sizes per axis, got {}'
                                 ''.format(discr_domain.partition.cell_sides))
            if geometry.ndim > 2 and impl.endswith('cpu'):
                raise ValueError('`impl` {}, only works for 2d geometries'
                                 ' got {}-d'.format(impl_in, geometry))
        elif impl == 'scikit':
            if not isinstance(geometry, Parallel2dGeometry):
                raise TypeError("'scikit' backend only supports 2d parallel "
                                'geometries')

            mid_pt = discr_domain.domain.mid_pt
            if not all(mid_pt == [0, 0]):
                raise ValueError('`discr_domain.domain` needs to be '
                                 'centered on [0, 0], got {}'.format(mid_pt))

            shape = discr_domain.shape
            if shape[0] != shape[1]:
                raise ValueError('`discr_domain.shape` needs to be square '
                                 'got {}'.format(shape))

            extent = discr_domain.domain.extent()
            if extent[0] != extent[1]:
                raise ValueError('`discr_domain.extent` needs to be square '
                                 'got {}'.format(extent))

        # TODO: sanity checks between domain and geometry (ndim, ...)
        self.__geometry = geometry
        self.__impl = impl
        self.kwargs = kwargs

        discr_range = kwargs.pop('discr_range', None)
        if discr_range is None:
            dtype = discr_domain.dspace.dtype

            # Create a discretized space (operator range) with the same
            # data-space type as the domain.
            # TODO: use a ProductSpace structure or find a way to treat
            # different dimensions differently in DiscreteLp
            # (i.e. in partitions).
            range_uspace = FunctionSpace(geometry.params,
                                         out_dtype=dtype)

            # Approximate cell volume
            # TODO: angles and detector must be handled separately. While the
            # detector should be uniformly discretized, the angles do not have
            # to and often are not.
            extent = float(geometry.partition.extent().prod())
            size = float(geometry.partition.size)
            weight = extent / size

            range_dspace = discr_domain.dspace_type(geometry.partition.size,
                                                    weight=weight, dtype=dtype)

            range_interp = kwargs.get('interp', 'nearest')
            discr_range = DiscreteLp(
                range_uspace, geometry.partition, range_dspace,
                interp=range_interp, order=discr_domain.order)

        super().__init__(discr_domain, discr_range, linear=True)

    @property
    def impl(self):
        """Implementation back-end for evaluation of this operator."""
        return self.__impl

    @property
    def geometry(self):
        """Geometry of this operator."""
        return self.__geometry

    def _call(self, x, out=None):
        """Forward project ``x`` and store the result in ``out`` if given."""
        if self.impl.startswith('astra'):
            backend, data_impl = self.impl.split('_')
            if data_impl == 'cpu':
                return astra_cpu_forward_projector(x, self.geometry,
                                                   self.range, out)
            elif data_impl == 'cuda':
                return astra_cuda_forward_projector(x, self.geometry,
                                                    self.range, out)
            else:
                # Should never happen
                raise RuntimeError('implementation info is inconsistent')
        elif self.impl == 'scikit':
            return scikit_radon_forward(x, self.geometry, self.range, out)
        else:  # Should never happen
            raise RuntimeError('implementation info is inconsistent')

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : `RayBackProjection`
        """
        kwargs = self.kwargs.copy()
        kwargs['discr_domain'] = self.range
        return RayBackProjection(self.domain, self.geometry, self.impl,
                                 **kwargs)


class RayBackProjection(Operator):
    """Adjoint of the discrete Ray transform between L^p spaces."""

    def __init__(self, discr_range, geometry, impl='astra_cpu', **kwargs):
        """Initialize a new instance.

        Parameters
        ----------
        discr_range : `DiscreteLp`
            Reconstruction space, the range of the back-projector
        geometry : `Geometry`
            The geometry of the transform, contains information about
            the operator domain
        impl : {'astra_cpu', 'astra_cuda', 'scikit'}, optional
            Implementation back-end for the transform. Supported back-ends:
            'astra_cpu': ASTRA toolbox using CPU, only 2D
            'astra_cuda': ASTRA toolbox, using CUDA, 2D or 3D
            'scikit': scikit-image, only 2D parallel with square domain
        interp : {'nearest', 'linear'}
            Interpolation type for the discretization of the operator range.
            Default: 'nearest'

        Other Parameters
        ----------------
        discr_domain : `DiscreteLp`
            Discretized space, the range of the forward projector.
            Default: Infered from parameters.
        """
        if not isinstance(discr_range, DiscreteLp):
            raise TypeError('`discr_range` {!r} is not a `DiscreteLp`'
                            ' instance'.format(discr_range))

        if not isinstance(geometry, Geometry):
            raise TypeError('`geometry` {!r} is not a `Geometry` instance'
                            ''.format(geometry))

        impl, impl_in = str(impl).lower(), impl
        if impl not in _SUPPORTED_IMPL:
            raise ValueError("`impl` '{}' not supported"
                             ''.format(impl_in))

        if impl.startswith('astra'):
            if not ASTRA_AVAILABLE:
                raise ValueError("'astra' backend not available")
            if impl == 'astra_cuda' and not ASTRA_CUDA_AVAILABLE:
                raise ValueError("'astra_cuda' backend not available")
            if not np.allclose(discr_range.partition.cell_sides[1:],
                               discr_range.partition.cell_sides[:-1]):
                raise ValueError('ASTRA does not support different voxel '
                                 'sizes per axis, got {}'
                                 ''.format(discr_range.partition.cell_sides))

        self.__geometry = geometry
        self.__impl = impl
        self.kwargs = kwargs

        discr_domain = kwargs.pop('discr_domain', None)
        if discr_domain is None:
            dtype = discr_range.dspace.dtype

            # Create a discretized space (operator domain) with the same
            # data-space type as the range.
            domain_uspace = FunctionSpace(geometry.params, out_dtype=dtype)

            # Approximate cell volume
            extent = float(geometry.partition.extent().prod())
            size = float(geometry.partition.size)
            weight = extent / size

            domain_dspace = discr_range.dspace_type(geometry.partition.size,
                                                    weight=weight, dtype=dtype)

            domain_interp = kwargs.get('interp', 'nearest')
            discr_domain = DiscreteLp(
                domain_uspace, geometry.partition, domain_dspace,
                interp=domain_interp, order=discr_range.order)
        super().__init__(discr_domain, discr_range, linear=True)

    @property
    def impl(self):
        """Implementation back-end for evaluation of this operator."""
        return self.__impl

    @property
    def geometry(self):
        """Geometry of this operator."""
        return self.__geometry

    def _call(self, x, out=None):
        """Back-project ``x`` and store the result in ``out`` if given."""

        if self.impl.startswith('astra'):
            backend, data_impl = self.impl.split('_')
            if data_impl == 'cpu':
                return astra_cpu_back_projector(x, self.geometry,
                                                self.range, out)
            elif data_impl == 'cuda':
                return astra_cuda_back_projector(x, self.geometry,
                                                 self.range, out)
            else:
                # Should never happen
                raise RuntimeError('implementation info is inconsistent')
        elif self.impl == 'scikit':
            return scikit_radon_back_projector(x, self.geometry,
                                               self.range, out)
        else:  # Should never happen
            raise RuntimeError('implementation info is inconsistent')

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : `RayTransform`
        """
        kwargs = self.kwargs.copy()
        kwargs['discr_domain'] = self.domain
        return RayTransform(self.range, self.geometry, impl=self.impl,
                            **kwargs)


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Default operators defined on any `ProductSpace`."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import super

import numpy as np
import scipy as sp
from numbers import Integral

from odl.operator.operator import Operator
from odl.space import ProductSpace


__all__ = ('ProductSpaceOperator',
           'ComponentProjection', 'ComponentProjectionAdjoint',
           'BroadcastOperator', 'ReductionOperator', 'DiagonalOperator')


class ProductSpaceOperator(Operator):

    """A "matrix of operators" on product spaces.

    For example a matrix of operators can act on a vector by

        ``ProductSpaceOperator([[A, B], [C, D]])([x, y]) =
        [A(x) + B(y), C(x) + D(y)]``

    Notes
    -----
    This is intended for the case where an operator can be decomposed
    as a linear combination of "sub-operators", e.g.

    .. math::

        \\left(
        \\begin{array}{ccc}
        A & B & 0 \\\\
        0 & C & 0 \\\\
        0 & 0 & D
        \end{array}\\right)
        \\left(
        \\begin{array}{c}
        x \\\\
        y \\\\
        z
        \end{array}\\right)
        =
        \\left(
        \\begin{array}{c}
        A(x) + B(y) \\\\
        C(y) \\\\
        D(z)
        \end{array}\\right)

    Mathematically, a `ProductSpaceOperator` is an operator

        :math:`\mathcal{A}: \mathcal{X} \\to \mathcal{Y}`

    between product spaces
    :math:`\mathcal{X}=\mathcal{X}_1 \\times\dots\\times \mathcal{X}_m`
    and
    :math:`\mathcal{Y}=\mathcal{Y}_1 \\times\dots\\times \mathcal{Y}_n`
    which can be written in the form

        :math:`\mathcal{A} = (\mathcal{A}_{ij})_{i,j},  \quad
        i = 1, \dots, n, \\ j = 1, \dots, m`

    with *component operators*
    :math:`\mathcal{A}_{ij}: \mathcal{X}_j \\to \mathcal{Y}_i`.

    Its action on a vector :math:`x = (x_1, \dots, x_m)` is defined as
    the matrix multiplication

        :math:`[\mathcal{A}(x)]_i = \sum_{j=1}^m \mathcal{A}_{ij}(x_j)`.

    See Also
    --------
    BroadcastOperator : Case when a single argument is used by several ops.
    ReductionOperator : Calculates sum of operator results.
    DiagonalOperator : Case where the 'matrix' is diagonal.
    """

    def __init__(self, operators, domain=None, range=None):
        """Initialize a new instance.

        Parameters
        ----------
        operators : `array-like`
            An array of `Operator`'s
        domain : `ProductSpace`, optional
            Domain of the operator. If not provided, it is tried to be
            inferred from the operators. This requires each **column**
            to contain at least one operator.
        range : `ProductSpace`, optional
            Range of the operator. If not provided, it is tried to be
            inferred from the operators. This requires each **row**
            to contain at least one operator.

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> X = odl.ProductSpace(r3, r3)
        >>> I = odl.IdentityOperator(r3)
        >>> x = X.element([[1, 2, 3], [4, 5, 6]])

        Sum of elements:

        >>> prod_op = ProductSpaceOperator([I, I])
        >>> prod_op(x)
        ProductSpace(rn(3), 1).element([
            [5.0, 7.0, 9.0]
        ])

        Diagonal operator -- 0 or ``None`` means ignore, or the implicit
        zero operator:

        >>> prod_op = ProductSpaceOperator([[I, 0], [0, I]])
        >>> prod_op(x)
        ProductSpace(rn(3), 2).element([
            [1.0, 2.0, 3.0],
            [4.0, 5.0, 6.0]
        ])

        Complicated combinations:

        >>> prod_op = ProductSpaceOperator([[I, I], [I, 0]])
        >>> prod_op(x)
        ProductSpace(rn(3), 2).element([
            [5.0, 7.0, 9.0],
            [1.0, 2.0, 3.0]
        ])
        """

        # Validate input data
        if domain is not None:
            if not isinstance(domain, ProductSpace):
                raise TypeError('`domain` {!r} not a ProductSpace instance'
                                ''.format(domain))
            if domain.is_weighted:
                raise NotImplementedError('weighted spaces not supported')

        if range is not None:
            if not isinstance(range, ProductSpace):
                raise TypeError('`range` {!r} not a ProductSpace instance'
                                ''.format(range))
            if range.is_weighted:
                raise NotImplementedError('weighted spaces not supported')

        # Convert ops to sparse representation
        self.ops = sp.sparse.coo_matrix(operators)

        if not all(isinstance(op, Operator) for op in self.ops.data):
            raise TypeError('`operators` {!r} must be a matrix of Operators'
                            ''.format(operators))

        # Set domain and range (or verify if given)
        if domain is None:
            domains = [None] * self.ops.shape[1]
        else:
            domains = domain

        if range is None:
            ranges = [None] * self.ops.shape[0]
        else:
            ranges = range

        for row, col, op in zip(self.ops.row, self.ops.col, self.ops.data):
            if domains[col] is None:
                domains[col] = op.domain
            elif domains[col] != op.domain:
                raise ValueError('column {}, has inconsistent domains, '
                                 'got {} and {}'
                                 ''.format(col, domains[col], op.domain))

            if ranges[row] is None:
                ranges[row] = op.range
            elif ranges[row] != op.range:
                raise ValueError('row {}, has inconsistent ranges, '
                                 'got {} and {}'
                                 ''.format(row, ranges[row], op.range))

        if domain is None:
            for col, sub_domain in enumerate(domains):
                if sub_domain is None:
                    raise ValueError('col {} empty, unable to determine '
                                     'domain, please use `domain` parameter'
                                     ''.format(col))

            domain = ProductSpace(*domains)

        if range is None:
            for row, sub_range in enumerate(ranges):
                if sub_range is None:
                    raise ValueError('row {} empty, unable to determine '
                                     'range, please use `range` parameter'
                                     ''.format(row))

            range = ProductSpace(*ranges)

        # Set linearity
        linear = all(op.is_linear for op in self.ops.data)

        super().__init__(domain=domain, range=range, linear=linear)

    def _call(self, x, out=None):
        """Call the operators on the parts of ``x``."""
        # TODO: add optimization in case an operator appears repeatedly in a
        # row
        if out is None:
            out = self.range.zero()
            for i, j, op in zip(self.ops.row, self.ops.col, self.ops.data):
                out[i] += op(x[j])
        else:
            has_evaluated_row = np.zeros(self.range.size, dtype=bool)
            for i, j, op in zip(self.ops.row, self.ops.col, self.ops.data):
                if not has_evaluated_row[i]:
                    op(x[j], out=out[i])
                else:
                    # TODO: optimize
                    out[i] += op(x[j])

                has_evaluated_row[i] = True

            for i, evaluated in enumerate(has_evaluated_row):
                if not evaluated:
                    out[i].set_zero()

        return out

    def derivative(self, x):
        """Derivative of the product space operator.

        Parameters
        ----------
        x : `domain` element
            The point to take the derivative in

        Returns
        -------
        adjoint : linear`ProductSpaceOperator`
            The derivative

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> X = odl.ProductSpace(r3, r3)
        >>> I = odl.IdentityOperator(r3)
        >>> x = X.element([[1, 2, 3], [4, 5, 6]])

        Example with linear operator (derivative is itself)

        >>> prod_op = ProductSpaceOperator([[0, I], [0, 0]],
        ...                                domain=X, range=X)
        >>> prod_op(x)
        ProductSpace(rn(3), 2).element([
            [4.0, 5.0, 6.0],
            [0.0, 0.0, 0.0]
        ])
        >>> prod_op.derivative(x)(x)
        ProductSpace(rn(3), 2).element([
            [4.0, 5.0, 6.0],
            [0.0, 0.0, 0.0]
        ])

        Example with affine operator

        >>> residual_op = I - r3.element([1, 1, 1])
        >>> op = ProductSpaceOperator([[0, residual_op], [0, 0]],
        ...                           domain=X, range=X)

        Calling operator gives offset by [1, 1, 1]

        >>> op(x)
        ProductSpace(rn(3), 2).element([
            [3.0, 4.0, 5.0],
            [0.0, 0.0, 0.0]
        ])

        Derivative of affine operator does not have this offset

        >>> op.derivative(x)(x)
        ProductSpace(rn(3), 2).element([
            [4.0, 5.0, 6.0],
            [0.0, 0.0, 0.0]
        ])
        """
        deriv_ops = [op.derivative(x[col]) for op, col in zip(self.ops.data,
                                                              self.ops.col)]
        indices = [self.ops.row, self.ops.col]
        shape = self.ops.shape
        deriv_matrix = sp.sparse.coo_matrix((deriv_ops, indices), shape)
        return ProductSpaceOperator(deriv_matrix, self.domain, self.range)

    @property
    def adjoint(self):
        """Adjoint of this operator.

        The adjoint is given by taking the transpose of the matrix
        and the adjoint of each component operator.

        In weighted product spaces, the adjoint needs to take the
        weightings into account. This is currently not supported.

        Returns
        -------
        adjoint : `ProductSpaceOperator`
            The adjoint

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> X = odl.ProductSpace(r3, r3)
        >>> I = odl.IdentityOperator(r3)
        >>> x = X.element([[1, 2, 3], [4, 5, 6]])

        Matrix is transposed:

        >>> prod_op = ProductSpaceOperator([[0, I], [0, 0]],
        ...                                domain=X, range=X)
        >>> prod_op(x)
        ProductSpace(rn(3), 2).element([
            [4.0, 5.0, 6.0],
            [0.0, 0.0, 0.0]
        ])
        >>> prod_op.adjoint(x)
        ProductSpace(rn(3), 2).element([
            [0.0, 0.0, 0.0],
            [1.0, 2.0, 3.0]
        ])
        """
        adjoint_ops = [op.adjoint for op in self.ops.data]
        indices = [self.ops.col, self.ops.row]  # Swap col/row -> transpose
        shape = (self.ops.shape[1], self.ops.shape[0])
        adj_matrix = sp.sparse.coo_matrix((adjoint_ops, indices), shape)
        return ProductSpaceOperator(adj_matrix, self.range, self.domain)

    def __repr__(self):
        """Return ``repr(self)``."""
        return 'ProductSpaceOperator({!r})'.format(self.ops)


class ComponentProjection(Operator):

    """Projection onto the subspace identified by an index.

    For a product space :math:`\mathcal{X} = \mathcal{X}_1 \\times \dots
    \\times \mathcal{X}_n`, the component projection

        :math:`\mathcal{P}_i: \mathcal{X} \\to \mathcal{X}_i`

    is given by :math:`\mathcal{P}_i(x) = x_i` for an element
    :math:`x = (x_1, \dots, x_n) \\in \mathcal{X}`.

    More generally, for an index set :math:`I \subset \{1, \dots, n\}`,
    the projection operator :math:`\mathcal{P}_I` is defined by
    :math:`\mathcal{P}_I(x) = (x_i)_{i \\in I}`.

    Note that this is a special case of a product space operator where
    the "operator matrix" has only one row and contains only
    identity operators.
    """

    def __init__(self, space, index):
        """Initialize a new instance.

        Parameters
        ----------
        space : `ProductSpace`
            Space to project from.
        index : int, slice, or iterable
            Indices defining the subspace. If ``index`` is not an integer,
            the `Operator.range` of this operator is also a `ProductSpace`.

        Examples
        --------
        >>> r1 = odl.rn(1)
        >>> r2 = odl.rn(2)
        >>> r3 = odl.rn(3)
        >>> X = odl.ProductSpace(r1, r2, r3)

        Projection on n-th component

        >>> proj = odl.ComponentProjection(X, 0)
        >>> x = [[1.0],
        ...      [2.0, 3.0],
        ...      [4.0, 5.0, 6.0]]
        >>> proj(x)
        rn(1).element([1.0])

        Projection on sub-space

        >>> proj = odl.ComponentProjection(X, [0, 2])
        >>> proj(x)
        ProductSpace(rn(1), rn(3)).element([
            [1.0],
            [4.0, 5.0, 6.0]
        ])
        """
        self.__index = index
        super().__init__(space, space[index], linear=True)

    @property
    def index(self):
        """Index of the subspace."""
        return self.__index

    def _call(self, x, out=None):
        """Project ``x`` onto the subspace."""
        if out is None:
            out = x[self.index].copy()
        else:
            out.assign(x[self.index])
        return out

    @property
    def adjoint(self):
        """Return the adjoint operator.

        The adjoint is given by extending along `ComponentProjection.index`,
        and setting zero along the others.

        See Also
        --------
        ComponentProjectionAdjoint
        """
        return ComponentProjectionAdjoint(self.domain, self.index)


class ComponentProjectionAdjoint(Operator):

    """Adjoint operator to `ComponentProjection`.

    As a special case of the adjoint of a `ProductSpaceOperator`,
    this operator is given as a column vector of identity operators
    and zero operators, with the identities placed in the positions
    defined by `ComponentProjectionAdjoint.index`.

    In weighted product spaces, the adjoint needs to take the
    weightings into account. This is currently not supported.
    """

    def __init__(self, space, index):
        """Initialize a new instance

        Parameters
        ----------
        space : `ProductSpace`
            Space to project to.
        index : int, slice, or iterable
            Indexes to project from.

        Examples
        --------
        >>> r1 = odl.rn(1)
        >>> r2 = odl.rn(2)
        >>> r3 = odl.rn(3)
        >>> X = odl.ProductSpace(r1, r2, r3)
        >>> x = X.element([[1], [2, 3], [4, 5, 6]])

        Projection on the 0-th component:

        >>> proj = odl.ComponentProjectionAdjoint(X, 0)
        >>> proj(x[0])
        ProductSpace(rn(1), rn(2), rn(3)).element([
            [1.0],
            [0.0, 0.0],
            [0.0, 0.0, 0.0]
        ])

        Projection on a sub-space corresponding to indices 0 and 2:

        >>> proj = odl.ComponentProjectionAdjoint(X, [0, 2])
        >>> proj(x[0, 2])
        ProductSpace(rn(1), rn(2), rn(3)).element([
            [1.0],
            [0.0, 0.0],
            [4.0, 5.0, 6.0]
        ])
        """
        self.__index = index
        super().__init__(space[index], space, linear=True)

    @property
    def index(self):
        """Index of the subspace."""
        return self.__index

    def _call(self, x, out=None):
        """Extend ``x`` from the subspace."""
        if out is None:
            out = self.range.zero()
        else:
            out.set_zero()

        out[self.index] = x
        return out

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : `ComponentProjection`
            The adjoint is given by the `ComponentProjection` related to this
            operator's `index`.
        """
        return ComponentProjection(self.range, self.index)


class BroadcastOperator(Operator):
    """Broadcast argument to set of operators.

    An argument is broadcast by evaluating several operators in the same
    point::

        BroadcastOperator(op1, op2)(x) = [op1(x), op2(x)]

    See Also
    --------
    ProductSpaceOperator : More general case, used as backend.
    ReductionOperator : Calculates sum of operator results.
    DiagonalOperator : Case where each operator should have its own argument.
    """
    def __init__(self, *operators):
        """Initialize a new instance

        Parameters
        ----------
        operator1,...,operatorN : `Operator` or `int`
            The individual operators that should be evaluated.
            Can also be given as ``operator, n`` with ``n`` integer,
            in which case ``operator`` is repeated ``n`` times.

        Examples
        --------
        Initialize an operator:

        >>> I = odl.IdentityOperator(odl.rn(3))
        >>> op = BroadcastOperator(I, 2 * I)
        >>> op.domain
        rn(3)
        >>> op.range
        ProductSpace(rn(3), 2)

        Evaluate the operator:

        >>> x = [1, 2, 3]
        >>> op(x)
        ProductSpace(rn(3), 2).element([
            [1.0, 2.0, 3.0],
            [2.0, 4.0, 6.0]
        ])

        Can also initialize by calling an operator repeatedly:

        >>> I = odl.IdentityOperator(odl.rn(3))
        >>> op = BroadcastOperator(I, 2)
        >>> op.operators
        (IdentityOperator(rn(3)), IdentityOperator(rn(3)))
        """
        if (len(operators) == 2 and
                isinstance(operators[0], Operator) and
                isinstance(operators[1], Integral)):
            operators = (operators[0],) * operators[1]

        self.__operators = operators
        self.__prod_op = ProductSpaceOperator([[op] for op in operators])

        super().__init__(self.prod_op.domain[0],
                         self.prod_op.range,
                         self.prod_op.is_linear)

    @property
    def prod_op(self):
        """`ProductSpaceOperator` implementation."""
        return self.__prod_op

    @property
    def operators(self):
        """Tuple of sub-operators that comprise ``self``."""
        return self.__operators

    def __getitem__(self, index):
        """Return an operator by index."""
        return self.operators[index]

    def _call(self, x, out=None):
        """Evaluate all operators in ``x`` and broadcast."""
        wrapped_x = self.prod_op.domain.element([x], cast=False)
        return self.prod_op(wrapped_x, out=out)

    def derivative(self, x):
        """Derivative of the broadcast operator.

        Parameters
        ----------
        x : `domain` element
            The point to take the derivative in

        Returns
        -------
        adjoint : linear `BroadcastOperator`
            The derivative

        Examples
        --------
        Example with an affine operator:

        >>> I = odl.IdentityOperator(odl.rn(3))
        >>> residual_op = I - I.domain.element([1, 1, 1])
        >>> op = BroadcastOperator(residual_op, 2 * residual_op)

        Calling operator offsets by ``[1, 1, 1]``:

        >>> x = [1, 2, 3]
        >>> op(x)
        ProductSpace(rn(3), 2).element([
            [0.0, 1.0, 2.0],
            [0.0, 2.0, 4.0]
        ])

        The derivative of this affine operator does not have an offset:

        >>> op.derivative(x)(x)
        ProductSpace(rn(3), 2).element([
            [1.0, 2.0, 3.0],
            [2.0, 4.0, 6.0]
        ])
        """
        return BroadcastOperator(*[op.derivative(x) for op in
                                   self.operators])

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : linear `BroadcastOperator`

        Examples
        --------
        >>> I = odl.IdentityOperator(odl.rn(3))
        >>> op = BroadcastOperator(I, 2 * I)
        >>> op.adjoint([[1, 2, 3], [2, 3, 4]])
        rn(3).element([5.0, 8.0, 11.0])
        """
        return ReductionOperator(*[op.adjoint for op in self.operators])


class ReductionOperator(Operator):
    """Reduce argument over set of operators.

    An argument is reduced by evaluating several operators and summing the
    result::

        ReductionOperator(op1, op2)(x) = op1(x[0]) + op2(x[1])

    See Also
    --------
    ProductSpaceOperator : More general case, used as backend.
    BroadcastOperator : Calls several operators with same argument.
    DiagonalOperator : Case where each operator should have its own argument.
    """
    def __init__(self, *operators):
        """Initialize a new instance.

        Parameters
        ----------
        operator1,...,operatorN : `Operator` or `int`
            The individual operators that should be evaluated and summed.
            Can also be given as ``operator, n`` with ``n`` integer,
            in which case ``operator`` is repeated ``n`` times.

        Examples
        --------
        >>> I = odl.IdentityOperator(odl.rn(3))
        >>> op = ReductionOperator(I, 2 * I)
        >>> op.domain
        ProductSpace(rn(3), 2)
        >>> op.range
        rn(3)

        Evaluating in a point gives sum:

        >>> op([[1.0, 2.0, 3.0],
        ...     [4.0, 6.0, 8.0]])
        rn(3).element([9.0, 14.0, 19.0])

        Can also be created using a multiple of a single operator:

        >>> op = ReductionOperator(I, 2)
        >>> op.operators
        (IdentityOperator(rn(3)), IdentityOperator(rn(3)))
        """
        if (len(operators) == 2 and
                isinstance(operators[0], Operator) and
                isinstance(operators[1], Integral)):
            operators = (operators[0],) * operators[1]

        self.__operators = operators
        self.__prod_op = ProductSpaceOperator([operators])

        super().__init__(self.prod_op.domain,
                         self.prod_op.range[0],
                         self.prod_op.is_linear)

    @property
    def prod_op(self):
        """`ProductSpaceOperator` implementation."""
        return self.__prod_op

    @property
    def operators(self):
        """Tuple of sub-operators that comprise ``self``."""
        return self.__operators

    def __getitem__(self, index):
        """Return an operator by index."""
        return self.operators[index]

    def _call(self, x, out=None):
        """Apply operators to ``x`` and sum."""
        if out is None:
            return self.prod_op(x)[0]
        else:
            wrapped_out = self.prod_op.range.element([out], cast=False)
            return self.prod_op(x, out=wrapped_out)

    def derivative(self, x):
        """Derivative of the reduction operator.

        Parameters
        ----------
        x : `domain` element
            The point to take the derivative in.

        Returns
        -------
        derivative : linear `BroadcastOperator`

        Examples
        --------
        >>> r3 = odl.rn(3)
        >>> I = odl.IdentityOperator(r3)
        >>> x = [1.0, 2.0, 3.0]
        >>> y = [4.0, 6.0, 8.0]

        Example with linear operator (derivative is itself)

        >>> op = ReductionOperator(I, 2 * I)
        >>> op([x, y])
        rn(3).element([9.0, 14.0, 19.0])
        >>> op.derivative([x, y])([x, y])
        rn(3).element([9.0, 14.0, 19.0])

        Example with affine operator

        >>> residual_op = I - r3.element([1, 1, 1])
        >>> op = ReductionOperator(residual_op, 2 * residual_op)

        Calling operator gives offset by [3, 3, 3]

        >>> op([x, y])
        rn(3).element([6.0, 11.0, 16.0])

        Derivative of affine operator does not have this offset

        >>> op.derivative([x, y])([x, y])
        rn(3).element([9.0, 14.0, 19.0])
        """
        return ReductionOperator(*[op.derivative(xi)
                                   for op, xi in zip(self.operators, x)])

    @property
    def adjoint(self):
        """Adjoint of this operator.

        Returns
        -------
        adjoint : linear `BroadcastOperator`

        Examples
        --------
        >>> I = odl.IdentityOperator(odl.rn(3))
        >>> op = ReductionOperator(I, 2 * I)
        >>> op.adjoint([1, 2, 3])
        ProductSpace(rn(3), 2).element([
            [1.0, 2.0, 3.0],
            [2.0, 4.0, 6.0]
        ])
        """
        return BroadcastOperator(*[op.adjoint for op in self.operators])


class DiagonalOperator(ProductSpaceOperator):
    """Diagonal 'matrix' of operators.

    For example, if ``A`` and ``B`` are operators, the diagonal operator
    can be seen as a matrix of operators::

        [[A, 0],
         [0, B]]

    When evaluated it gives::

         DiagonalOperator(op1, op2)(x) = [op1(x), op2(x)]

    See Also
    --------
    ProductSpaceOperator : Case when the 'matrix' is dense.
    BroadcastOperator : Case when a single argument is used by several ops.
    ReductionOperator : Calculates sum of operator results.
    """

    def __init__(self, *operators, **kwargs):
        """Initialize a new instance.

        Parameters
        ----------
        operator1,...,operatorN : `Operator` or int
            The individual operators in the diagonal.
            Can be specified as ``operator, n`` with ``n`` integer,
            in which case the diagonal operator with ``n`` multiples of
            ``operator`` is created.
        kwargs :
            Keyword arguments passed to the `ProductSpaceOperator` backend.

        Examples
        --------
        >>> I = odl.IdentityOperator(odl.rn(3))
        >>> op = DiagonalOperator(I, 2 * I)
        >>> op.domain
        ProductSpace(rn(3), 2)
        >>> op.range
        ProductSpace(rn(3), 2)

        Evaluation is distributed so each argument is given to one operator.
        The argument order is the same as the order of the operators:

        >>> op([[1, 2, 3],
        ...     [4, 5, 6]])
        ProductSpace(rn(3), 2).element([
            [1.0, 2.0, 3.0],
            [8.0, 10.0, 12.0]
        ])

        Can also be created using a multiple of a single operator

        >>> op = DiagonalOperator(I, 2)
        >>> op.operators
        (IdentityOperator(rn(3)), IdentityOperator(rn(3)))
        """
        if (len(operators) == 2 and
                isinstance(operators[0], Operator) and
                isinstance(operators[1], Integral)):
            operators = (operators[0],) * operators[1]

        indices = [range(len(operators)), range(len(operators))]
        shape = (len(operators), len(operators))
        op_matrix = sp.sparse.coo_matrix((operators, indices), shape)

        self.__operators = tuple(operators)

        super().__init__(op_matrix, **kwargs)

    @property
    def operators(self):
        """Tuple of sub-operators that comprise ``self``."""
        return self.__operators

    def __getitem__(self, index):
        """Return an operator by index."""
        return self.operators[index]

    def derivative(self, point):
        """Derivative of this operator.

        For example, if A and B are operators

            [[A, 0],
             [0, B]]

        The derivative is given by:

            [[A', 0],
             [0, B']]

        This is only well defined if each sub-operator has a derivative

        Parameters
        ----------
        point : `element-like` in ``domain``
            The point in which the derivative should be taken.

        Returns
        -------
        derivative : `DiagonalOperator`
            The derivative operator

        See Also
        --------
        ProductSpaceOperator.derivative
        """
        point = self.domain.element(point)

        derivs = [op.derivative(p) for op, p in zip(self.operators, point)]
        return DiagonalOperator(*derivs,
                                domain=self.domain, range=self.range)

    @property
    def adjoint(self):
        """Adjoint of this operator.

        For example, if A and B are operators::

            [[A, 0],
             [0, B]]

        The adjoint is given by::

            [[A^*, 0],
             [0, B^*]]

        This is only well defined if each sub-operator has an adjoint

        Returns
        -------
        adjoint : `DiagonalOperator`
            The adjoint operator

        See Also
        --------
        ProductSpaceOperator.adjoint
        """
        adjoints = [op.adjoint for op in self.operators]
        return DiagonalOperator(*adjoints,
                                domain=self.range, range=self.domain)

    @property
    def inverse(self):
        """Inverse of this operator.

        For example, if A and B are operators::

            [[A, 0],
             [0, B]]

        The inverse is given by::

            [[A^-1, 0],
             [0, B^-1]]

        This is only well defined if each sub-operator has an inverse

        Returns
        -------
        inverse : `DiagonalOperator`
            The inverse operator

        See Also
        --------
        ProductSpaceOperator.inverse
        """
        inverses = [op.inverse for op in self.operators]
        return DiagonalOperator(*inverses,
                                domain=self.range, range=self.domain)


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.


"""ODL is a functional analysis library with a focus on discretization.

ODL suppors abstract sets, linear vector spaces defined on such
and Operators/Functionals defined on these sets. It is intended
to be used to write general code and faciliate code reuse.
"""

from __future__ import absolute_import

__version__ = '0.5.0'
__all__ = ('diagnostics', 'discr', 'operator', 'set', 'space', 'solvers',
           'tomo', 'trafos', 'util', 'phantom', 'deform')

# Propagate names defined in __all__ of all submodules into the top-level
# module
from . import diagnostics

from .discr import *
__all__ += discr.__all__

from .operator import *
__all__ += operator.__all__

from .set import *
__all__ += set.__all__

from .space import *
__all__ += space.__all__

from . import solvers
from . import trafos
from . import tomo
from . import util
from . import phantom
from . import deform

from .util import test
__all__ += ('test',)



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Spaces of functions with common domain and range."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import super

from inspect import isfunction
import numpy as np

from odl.operator.operator import Operator, _dispatch_call_args
from odl.set import (RealNumbers, ComplexNumbers, Set, Field, LinearSpace,
                     LinearSpaceElement)
from odl.util.utility import (is_real_dtype, is_complex_floating_dtype,
                              preload_first_arg, dtype_repr,
                              complex_dtype, real_dtype)
from odl.util.vectorization import (
    is_valid_input_array, is_valid_input_meshgrid,
    out_shape_from_array, out_shape_from_meshgrid, vectorize)


__all__ = ('FunctionSet', 'FunctionSetElement',
           'FunctionSpace', 'FunctionSpaceElement')


def _default_in_place(func, x, out, **kwargs):
    """Default in-place evaluation method."""
    out[:] = func(x, **kwargs)
    return out


def _default_out_of_place(func, x, **kwargs):
    """Default in-place evaluation method."""
    if is_valid_input_array(x, func.domain.ndim):
        out_shape = out_shape_from_array(x)
    elif is_valid_input_meshgrid(x, func.domain.ndim):
        out_shape = out_shape_from_meshgrid(x)
    else:
        raise TypeError('cannot use in-place method to implement '
                        'out-of-place non-vectorized evaluation')

    dtype = func.space.out_dtype
    if dtype is None:
        dtype = np.result_type(*x)

    out = np.empty(out_shape, dtype=dtype)
    func(x, out=out, **kwargs)
    return out


def _broadcast_to(array, shape):
    """Wrapper for the numpy function broadcast_to.

    Added since we dont require numpy 1.10 and hence cant guarantee that this
    exists.
    """
    array = np.asarray(array)
    try:
        return np.broadcast_to(array, shape)
    except AttributeError:
        # The above requires numpy 1.10, fallback impl else
        shape = [m if n == 1 and m != 1 else 1
                 for n, m in zip(array.shape, shape)]
        return array + np.zeros(shape, dtype=array.dtype)


class FunctionSet(Set):

    """A general set of functions with common domain and range."""

    def __init__(self, domain, range, out_dtype=None):
        """Initialize a new instance.

        Parameters
        ----------
        domain : `Set`
            The domain of the functions.
        range : `Set`
            The range of the functions.
        out_dtype : optional
            Data type of the return value of a function in this space.
            Can be given in any way `numpy.dtype` understands, e.g. as
            string ('bool') or data type (bool).
            If no data type is given, a "lazy" evaluation is applied,
            i.e. an adequate data type is inferred during function
            evaluation.
        """
        if not isinstance(domain, Set):
            raise TypeError('`domain` {!r} not a `Set` instance'
                            ''.format(domain))

        if not isinstance(range, Set):
            raise TypeError('`range` {!r} not a `Set` instance'
                            ''.format(range))

        self.__domain = domain
        self.__range = range
        self.__out_dtype = None if out_dtype is None else np.dtype(out_dtype)

    @property
    def domain(self):
        """Common domain of all functions in this set."""
        return self.__domain

    @property
    def range(self):
        """Common range of all functions in this set."""
        return self.__range

    @property
    def out_dtype(self):
        """Output data type of this function.

        If ``None``, the output data type is not uniquely pre-defined.
        """
        return self.__out_dtype

    def element(self, fcall=None, vectorized=True):
        """Create a `FunctionSet` element.

        Parameters
        ----------
        fcall : callable, optional
            The actual instruction for out-of-place evaluation.
            It must return a `FunctionSet.range` element or a
            `numpy.ndarray` of such (vectorized call).

        vectorized : bool
            Whether ``fcall`` supports vectorized evaluation.

        Returns
        -------
        element : `FunctionSetElement`
            The new element, always supports vectorization

        See Also
        --------
        odl.discr.grid.TensorGrid.meshgrid : efficient grids for function
            evaluation
        """
        if not callable(fcall):
            raise TypeError('`fcall` {!r} is not callable'.format(fcall))
        elif fcall in self:
            return fcall
        else:
            if not vectorized:
                fcall = vectorize(fcall)

            return self.element_type(self, fcall)

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if ``other`` is a `FunctionSet` with same
            `FunctionSet.domain` and `FunctionSet.range`, ``False`` otherwise.
        """
        if other is self:
            return True

        return (isinstance(other, FunctionSet) and
                self.domain == other.domain and
                self.range == other.range and
                self.out_dtype == other.out_dtype)

    def __contains__(self, other):
        """Return ``other in self``.

        Returns
        -------
        equals : bool
            ``True`` if ``other`` is a `FunctionSetElement`
            whose `FunctionSetElement.space` attribute
            equals this space, ``False`` otherwise.
        """
        return (isinstance(other, self.element_type) and
                self == other.space)

    def __repr__(self):
        """Return ``repr(self)``."""
        return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                       self.domain, self.range)

    def __str__(self):
        """Return ``str(self)``."""
        return '{}({}, {})'.format(self.__class__.__name__,
                                   self.domain, self.range)

    @property
    def element_type(self):
        """`FunctionSetElement`"""
        return FunctionSetElement


class FunctionSetElement(Operator):

    """Representation of a `FunctionSet` element."""

    def __init__(self, fset, fcall):
        """Initialize a new instance.

        Parameters
        ----------
        fset : `FunctionSet`
            Set of functions this element lives in.
        fcall : callable
            The actual instruction for out-of-place evaluation.
            It must return a `FunctionSet.range` element or a
            `numpy.ndarray` of such (vectorized call).
        """
        self.__space = fset
        super().__init__(self.space.domain, self.space.range, linear=False)

        # Determine which type of implementation fcall is
        if isinstance(fcall, FunctionSetElement):
            call_has_out, call_out_optional, _ = _dispatch_call_args(
                bound_call=fcall._call)

        # Numpy Ufuncs and similar objects (e.g. Numba DUfuncs)
        elif hasattr(fcall, 'nin') and hasattr(fcall, 'nout'):
            if fcall.nin != 1:
                raise ValueError('ufunc {} has {} input parameter(s), '
                                 'expected 1'
                                 ''.format(fcall.__name__, fcall.nin))
            if fcall.nout > 1:
                raise ValueError('ufunc {} has {} output parameter(s), '
                                 'expected at most 1'
                                 ''.format(fcall.__name__, fcall.nout))
            call_has_out = call_out_optional = (fcall.nout == 1)
        elif isfunction(fcall):
            call_has_out, call_out_optional, _ = _dispatch_call_args(
                unbound_call=fcall)
        elif callable(fcall):
            call_has_out, call_out_optional, _ = _dispatch_call_args(
                bound_call=fcall.__call__)
        else:
            raise TypeError('type {!r} not callable')

        self._call_has_out = call_has_out
        self._call_out_optional = call_out_optional

        if not call_has_out:
            # Out-of-place-only
            self._call_in_place = preload_first_arg(self, 'in-place')(
                _default_in_place)
            self._call_out_of_place = fcall
        elif call_out_optional:
            # Dual-use
            self._call_in_place = self._call_out_of_place = fcall
        else:
            # In-place-only
            self._call_in_place = fcall
            self._call_out_of_place = preload_first_arg(self, 'out-of-place')(
                _default_out_of_place)

    @property
    def space(self):
        """Space or set this function belongs to."""
        return self.__space

    @property
    def out_dtype(self):
        """Output data type of this function.

        If ``None``, the output data type is not uniquely pre-defined.
        """
        return self.space.out_dtype

    def _call(self, x, out=None, **kwargs):
        """Raw evaluation method."""
        if out is None:
            return self._call_out_of_place(x, **kwargs)
        else:
            self._call_in_place(x, out=out, **kwargs)

    def __call__(self, x, out=None, **kwargs):
        """Return ``self(x[, out, **kwargs])``.

        Parameters
        ----------
        x : `domain` `element-like`, `meshgrid` or `numpy.ndarray`
            Input argument for the function evaluation. Conditions
            on ``x`` depend on its type:

            element-like: must be a castable to a domain element

            meshgrid: length must be ``space.ndim``, and the arrays must
            be broadcastable against each other.

            array:  shape must be ``(d, N)``, where ``d`` is the number
            of dimensions of the function domain

        out : `numpy.ndarray`, optional
            Output argument holding the result of the function
            evaluation, can only be used for vectorized
            functions. Its shape must be equal to
            ``np.broadcast(*x).shape``.

        Other Parameters
        ----------------
        bounds_check : bool
            If ``True``, check if all input points lie in the function
            domain in the case of vectorized evaluation. This requires
            the domain to implement `Set.contains_all`.
            Default: ``True``

        Returns
        -------
        out : range element or array of elements
            Result of the function evaluation. If ``out`` was provided,
            the returned object is a reference to it.

        Raises
        ------
        TypeError
            If ``x`` is not a valid vectorized evaluation argument

            If ``out`` is not a range element or a `numpy.ndarray`
            of range elements

        ValueError
            If evaluation points fall outside the valid domain
        """
        bounds_check = kwargs.pop('bounds_check', True)
        if bounds_check and not hasattr(self.domain, 'contains_all'):
            raise AttributeError('bounds check not possible for '
                                 'domain {}, missing `contains_all()` '
                                 'method'.format(self.domain))

        if bounds_check and not hasattr(self.range, 'contains_all'):
            raise AttributeError('bounds check not possible for '
                                 'range {}, missing `contains_all()` '
                                 'method'.format(self.range))

        ndim = getattr(self.domain, 'ndim', None)
        # Check for input type and determine output shape
        if is_valid_input_meshgrid(x, ndim):
            out_shape = out_shape_from_meshgrid(x)
            scalar_out = False
            # Avoid operations on tuples like x * 2 by casting to array
            if ndim == 1:
                x = x[0][None, ...]
        elif is_valid_input_array(x, ndim):
            x = np.asarray(x)
            out_shape = out_shape_from_array(x)
            scalar_out = False
            # For 1d, squeeze the array
            if ndim == 1 and x.ndim == 2:
                x = x.squeeze()
        elif x in self.domain:
            x = np.atleast_2d(x).T  # make a (d, 1) array
            out_shape = (1,)
            scalar_out = (out is None)
        else:
            # Unknown input
            txt_1d = ' or (n,)' if ndim == 1 else ''
            raise TypeError('Argument {!r} not a valid vectorized '
                            'input. Expected an element of the domain '
                            '{domain}, an array-like with shape '
                            '({domain.ndim}, n){} or a length-{domain.ndim} '
                            'meshgrid tuple.'
                            ''.format(x, txt_1d, domain=self.domain))

        # Check bounds if specified
        if bounds_check:
            if not self.domain.contains_all(x):
                raise ValueError('input contains points outside '
                                 'the domain {}'.format(self.domain))

        # Call the function and check out shape, before or after
        if out is None:
            if ndim == 1:
                try:
                    out = self._call(x, **kwargs)
                except (TypeError, IndexError):
                    # TypeError is raised if a meshgrid was used but the
                    # function expected an array (1d only). In this case we try
                    # again with the first meshgrid vector.
                    # IndexError is raised in expressions like x[x > 0] since
                    # "x > 0" evaluates to 'True', i.e. 1, and that index is
                    # out of range for a meshgrid tuple of length 1 :-). To get
                    # the real errors with indexing, we check again for the
                    # same scenario (scalar output when not valid) as in the
                    # first case.
                    out = self._call(x[0], **kwargs)

                # squeeze to remove extra axes.
                out = np.squeeze(out)
            else:
                out = self._call(x, **kwargs)

            # Cast to proper dtype if needed, also convert to array if out
            # is scalar.
            out = np.asarray(out, self.out_dtype)

            if out_shape != (1,) and out.shape != out_shape:
                # Try to broadcast the returned element if possible
                out = _broadcast_to(out, out_shape)
        else:
            if not isinstance(out, np.ndarray):
                raise TypeError('output {!r} not a `numpy.ndarray` '
                                'instance')
            if out_shape != (1,) and out.shape != out_shape:
                raise ValueError('output shape {} not equal to shape '
                                 '{} expected from input'
                                 ''.format(out.shape, out_shape))
            if self.out_dtype is not None and out.dtype != self.out_dtype:
                raise ValueError('`out.dtype` ({}) does not match out_dtype '
                                 '({})'.format(out.dtype, self.out_dtype))

            if ndim == 1:
                # TypeError for meshgrid in 1d, but expected array (see above)
                try:
                    self._call(x, out=out, **kwargs)
                except TypeError:
                    self._call(x[0], out=out, **kwargs)
            else:
                self._call(x, out=out, **kwargs)

        # Check output values
        if bounds_check:
            if not self.range.contains_all(out):
                raise ValueError('output contains points outside '
                                 'the range {}'
                                 ''.format(self.range))

        # Numpy does not implement __complex__ for arrays (in contrast to
        # __float__), so we have to fish out the scalar ourselves.
        return self.range.element(out.ravel()[0]) if scalar_out else out

    def assign(self, other):
        """Assign ``other`` to ``self``.

        This is implemented without `FunctionSpace.lincomb` to ensure that
        ``self == other`` evaluates to True after ``self.assign(other)``.
        """
        if other not in self.space:
            raise TypeError('`other` {!r} is not an element of the space '
                            '{} of this function'
                            ''.format(other, self.space))
        self._call_in_place = other._call_in_place
        self._call_out_of_place = other._call_out_of_place
        self._call_has_out = other._call_has_out
        self._call_out_optional = other._call_out_optional

    def copy(self):
        """Create an identical (deep) copy of this element."""
        result = self.space.element()
        result.assign(self)
        return result

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if ``other`` is a `FunctionSetElement` with
            ``other.space == self.space``, and the functions for evaluation
            evaluation of ``self`` and ``other`` are the same, ``False``
            otherwise.
        """
        if other is self:
            return True

        if not isinstance(other, FunctionSetElement):
            return False

        # We cannot blindly compare since functions may have been wrapped
        if (self._call_has_out != other._call_has_out or
                self._call_out_optional != other._call_out_optional):
            return False

        if self._call_has_out:
            # Out-of-place can be wrapped in this case, so we compare only
            # the in-place methods.
            funcs_equal = self._call_in_place == other._call_in_place
        else:
            # Just the opposite of the first case
            funcs_equal = self._call_out_of_place == other._call_out_of_place

        return self.space == other.space and funcs_equal

    def __str__(self):
        """Return ``str(self)``."""
        if self._call_has_out:
            func = self._call_in_place
        else:
            func = self._call_out_of_place
        return '{}: {} --> {}'.format(func, self.domain, self.range)

    def __repr__(self):
        """Return ``repr(self)``."""
        if self._call_has_out:
            func = self._call_in_place
        else:
            func = self._call_out_of_place

        return '{!r}.element({!r})'.format(self.space, func)


class FunctionSpace(FunctionSet, LinearSpace):

    """A vector space of functions."""

    def __init__(self, domain, field=None, out_dtype=None):
        """Initialize a new instance.

        Parameters
        ----------
        domain : `Set`
            The domain of the functions
        field : `Field`, optional
            The range of the functions, usually the `RealNumbers` or
            `ComplexNumbers`. If not given, the field is either inferred
            from ``out_dtype``, or, if the latter is also ``None``, set
            to ``RealNumbers()``.
        out_dtype : optional
            Data type of the return value of a function in this space.
            Can be given in any way `numpy.dtype` understands, e.g. as
            string (``'float64'``) or data type (``float``).
            By default, ``'float64'`` is used for real and ``'complex128'``
            for complex spaces.
        """
        if not isinstance(domain, Set):
            raise TypeError('`domain` {!r} not a Set instance'.format(domain))

        if field is not None and not isinstance(field, Field):
            raise TypeError('`field` {!r} not a `Field` instance'
                            ''.format(field))

        # Data type: check if consistent with field, take default for None
        dtype, dtype_in = np.dtype(out_dtype), out_dtype

        # Default for both None
        if field is None and out_dtype is None:
            field = RealNumbers()
            out_dtype = np.dtype('float64')

        # field None, dtype given -> infer field
        elif field is None:
            if is_real_dtype(dtype):
                field = RealNumbers()
            elif is_complex_floating_dtype(dtype):
                field = ComplexNumbers()
            else:
                raise ValueError('{} is not a scalar data type'
                                 ''.format(dtype_in))

        # field given -> infer dtype if not given, else check consistency
        elif field == RealNumbers():
            if out_dtype is None:
                out_dtype = np.dtype('float64')
            elif not is_real_dtype(dtype):
                raise ValueError('{} is not a real data type'
                                 ''.format(dtype_in))
        elif field == ComplexNumbers():
            if out_dtype is None:
                out_dtype = np.dtype('complex128')
            elif not is_complex_floating_dtype(dtype):
                raise ValueError('{} is not a complex data type'
                                 ''.format(dtype_in))

        # Else: keep out_dtype=None, which results in lazy dtype determination

        LinearSpace.__init__(self, field)
        FunctionSet.__init__(self, domain, field, out_dtype)

        # Init cache attributes for real / complex variants
        if self.field == RealNumbers():
            self.__real_out_dtype = self.out_dtype
            self.__real_space = self
            self.__complex_out_dtype = complex_dtype(self.out_dtype,
                                                     default=np.dtype(object))
            self.__complex_space = None
        elif self.field == ComplexNumbers():
            self.__real_out_dtype = real_dtype(self.out_dtype)
            self.__real_space = None
            self.__complex_out_dtype = self.out_dtype
            self.__complex_space = self
        else:
            self.__real_out_dtype = None
            self.__real_space = None
            self.__complex_out_dtype = None
            self.__complex_space = None

    @property
    def real_out_dtype(self):
        """The real dtype corresponding to this space's `out_dtype`."""
        return self.__real_out_dtype

    @property
    def complex_out_dtype(self):
        """The complex dtype corresponding to this space's `out_dtype`."""
        return self.__complex_out_dtype

    @property
    def real_space(self):
        """The space corresponding to this space's `real_dtype`."""
        return self.astype(self.real_out_dtype)

    @property
    def complex_space(self):
        """The space corresponding to this space's `complex_dtype`."""
        return self.astype(self.complex_out_dtype)

    def element(self, fcall=None, vectorized=True):
        """Create a `FunctionSpace` element.

        Parameters
        ----------
        fcall : callable, optional
            The actual instruction for out-of-place evaluation.
            It must return a `FunctionSet.range` element or a
            `numpy.ndarray` of such (vectorized call).

            If fcall is a `FunctionSetElement`, it is wrapped
            as a new `FunctionSpaceElement`.

        vectorized : bool
            Whether ``fcall`` supports vectorized evaluation.

        Returns
        -------
        element : `FunctionSpaceElement`
            The new element, always supports vectorization

        Notes
        -----
        If you specify ``vectorized=False``, the function is decorated
        with a vectorizer, which makes two elements created this way
        from the same function being regarded as *not equal*.
        """
        if fcall is None:
            return self.zero()
        elif fcall in self:
            return fcall
        else:
            if not callable(fcall):
                raise TypeError('`fcall` {!r} is not callable'.format(fcall))
            if not vectorized:
                if self.field == RealNumbers():
                    dtype = 'float64'
                else:
                    dtype = 'complex128'

                fcall = vectorize(otypes=[dtype])(fcall)

            return self.element_type(self, fcall)

    def zero(self):
        """Function mapping everything to zero.

        This function is the additive unit in the function space.

        Since `FunctionSpace.lincomb` may be slow, we implement this function
        directly.
        """
        def zero_vec(x, out=None):
            """Zero function, vectorized."""
            if is_valid_input_meshgrid(x, self.domain.ndim):
                out_shape = out_shape_from_meshgrid(x)
            elif is_valid_input_array(x, self.domain.ndim):
                out_shape = out_shape_from_array(x)
            else:
                raise TypeError('invalid input type')

            if out is None:
                return np.zeros(out_shape, dtype=self.out_dtype)
            else:
                out.fill(0)

        return self.element_type(self, zero_vec)

    def one(self):
        """Function mapping everything to one.

        This function is the multiplicative unit in the function space.
        """
        def one_vec(x, out=None):
            """One function, vectorized."""
            if is_valid_input_meshgrid(x, self.domain.ndim):
                out_shape = out_shape_from_meshgrid(x)
            elif is_valid_input_array(x, self.domain.ndim):
                out_shape = out_shape_from_array(x)
            else:
                raise TypeError('invalid input type')

            if out is None:
                return np.ones(out_shape, dtype=self.out_dtype)
            else:
                out.fill(1)

        return self.element_type(self, one_vec)

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if ``other`` is a `FunctionSpace` with same
            `FunctionSpace.domain` and `FunctionSpace.range`,
            ``False`` otherwise.
        """
        if other is self:
            return True

        return (isinstance(other, FunctionSpace) and
                FunctionSet.__eq__(self, other))

    def _astype(self, out_dtype):
        """Internal helper for ``astype``."""
        return type(self)(self.domain, out_dtype=out_dtype)

    def astype(self, out_dtype):
        """Return a copy of this space with new ``out_dtype``.

        Parameters
        ----------
        out_dtype : optional
            Output data type of the returned space. Can be given in any
            way `numpy.dtype` understands, e.g. as string ('complex64')
            or data type (complex). None is interpreted as 'float64'.

        Returns
        -------
        newspace : `FunctionSpace`
            The version of this space with given data type
        """
        out_dtype = np.dtype(out_dtype)
        if out_dtype == self.out_dtype:
            return self

        # Caching for real and complex versions (exact dtyoe mappings)
        if out_dtype == self.real_out_dtype:
            if self.__real_space is None:
                self.__real_space = self._astype(out_dtype)
            return self.__real_space
        elif out_dtype == self.complex_out_dtype:
            if self.__complex_space is None:
                self.__complex_space = self._astype(out_dtype)
            return self.__complex_space
        else:
            return self._astype(out_dtype)

    def _lincomb(self, a, x1, b, x2, out):
        """Raw linear combination of ``x1`` and ``x2``.

        Notes
        -----
        The additions and multiplications are implemented via simple
        Python functions, so non-vectorized versions are slow.
        """
        # Store to allow aliasing
        x1_call_oop = x1._call_out_of_place
        x1_call_ip = x1._call_in_place
        x2_call_oop = x2._call_out_of_place
        x2_call_ip = x2._call_in_place

        def lincomb_call_out_of_place(x):
            """Linear combination, out-of-place version."""
            # Due to vectorization, at least one call must be made to
            # ensure the correct final shape. The rest is optimized as
            # far as possible.
            if a == 0 and b != 0:
                out = np.asarray(x2_call_oop(x), dtype=self.out_dtype)
                if b != 1:
                    out *= b
            elif b == 0:  # Contains the case a == 0
                out = np.asarray(x1_call_oop(x), dtype=self.out_dtype)
                if a != 1:
                    out *= a
            else:
                out = np.asarray(x1_call_oop(x), dtype=self.out_dtype)
                if a != 1:
                    out *= a
                tmp = np.asarray(x2_call_oop(x), dtype=self.out_dtype)
                if b != 1:
                    tmp *= b
                out += tmp
            return out

        def lincomb_call_in_place(x, out):
            """Linear combination, in-place version."""
            if not isinstance(out, np.ndarray):
                raise TypeError('in-place evaluation only possible if output '
                                'is of type `numpy.ndarray`')
            # TODO: this could be optimized for the case when x1 and x2
            # are identical
            if a == 0 and b == 0:
                out *= 0
            elif a == 0 and b != 0:
                x2_call_ip(x, out)
                if b != 1:
                    out *= b
            elif b == 0 and a != 0:
                x1_call_ip(x, out)
                if a != 1:
                    out *= a
            else:
                tmp = np.empty_like(out)
                x1_call_ip(x, out)
                x2_call_ip(x, tmp)
                if a != 1:
                    out *= a
                if b != 1:
                    tmp *= b
                out += tmp
            return out

        out._call_out_of_place = lincomb_call_out_of_place
        out._call_in_place = lincomb_call_in_place
        out._call_has_out = out._call_out_optional = True
        return out

    def _multiply(self, x1, x2, out):
        """Raw pointwise multiplication of two functions.

        Notes
        -----
        The multiplication is implemented with a simple Python
        function, so the non-vectorized versions are slow.
        """
        # Store to allow aliasing
        x1_call_oop = x1._call_out_of_place
        x1_call_ip = x1._call_in_place
        x2_call_oop = x2._call_out_of_place
        x2_call_ip = x2._call_in_place

        def product_call_out_of_place(x):
            """Product out-of-place evaluation function."""
            return np.asarray(x1_call_oop(x) * x2_call_oop(x),
                              dtype=self.out_dtype)

        def product_call_in_place(x, out):
            """Product in-place evaluation function."""
            tmp = np.empty_like(out, dtype=self.out_dtype)
            x1_call_ip(x, out)
            x2_call_ip(x, tmp)
            out *= tmp
            return out

        out._call_out_of_place = product_call_out_of_place
        out._call_in_place = product_call_in_place
        out._call_has_out = out._call_out_optional = True
        return out

    def _divide(self, x1, x2, out):
        """Raw pointwise division of two functions."""
        # Store to allow aliasing
        x1_call_oop = x1._call_out_of_place
        x1_call_ip = x1._call_in_place
        x2_call_oop = x2._call_out_of_place
        x2_call_ip = x2._call_in_place

        def quotient_call_out_of_place(x):
            """Quotient out-of-place evaluation function."""
            return np.asarray(x1_call_oop(x) / x2_call_oop(x),
                              dtype=self.out_dtype)

        def quotient_call_in_place(x, out):
            """Quotient in-place evaluation function."""
            tmp = np.empty_like(out, dtype=self.out_dtype)
            x1_call_ip(x, out)
            x2_call_ip(x, tmp)
            out /= tmp
            return out

        out._call_out_of_place = quotient_call_out_of_place
        out._call_in_place = quotient_call_in_place
        out._call_has_out = out._call_out_optional = True
        return out

    def _scalar_power(self, x, p, out):
        """Raw p-th power of a function, p integer or general scalar."""
        x_call_oop = x._call_out_of_place
        x_call_ip = x._call_in_place

        def pow_posint(x, n):
            """Recursion to calculate the n-th power out-of-place."""
            if isinstance(x, np.ndarray):
                y = x.copy()
                return ipow_posint(y, n)
            else:
                return x ** n

        def ipow_posint(x, n):
            """Recursion to calculate the n-th power in-place."""
            if n == 1:
                return x
            elif n % 2 == 0:
                x *= x
                return ipow_posint(x, n // 2)
            else:
                tmp = x.copy()
                x *= x
                ipow_posint(x, n // 2)
                x *= tmp
                return x

        def power_call_out_of_place(x):
            """Power out-of-place evaluation function."""
            if p == 0:
                return self.one()
            elif p == int(p) and p >= 1:
                return np.asarray(pow_posint(x_call_oop(x), int(p)),
                                  dtype=self.out_dtype)
            else:
                return np.power(x_call_oop(x), p).astype(self.out_dtype)

        def power_call_in_place(x, out):
            """Power in-place evaluation function."""
            if p == 0:
                out.assign(self.one())

            x_call_ip(x, out)
            if p == int(p) and p >= 1:
                return ipow_posint(out, int(p))
            else:
                out **= p
                return out

        out._call_out_of_place = power_call_out_of_place
        out._call_in_place = power_call_in_place
        out._call_has_out = out._call_out_optional = True
        return out

    def _realpart(self, x):
        """Function returning the real part of a result."""
        x_call_oop = x._call_out_of_place

        def realpart_oop(x):
            return np.asarray(x_call_oop(x), dtype=self.out_dtype).real

        if is_real_dtype(self.out_dtype):
            return x
        else:
            rdtype = real_dtype(self.out_dtype)
            rspace = self.astype(rdtype)
            return rspace.element(realpart_oop)

    def _imagpart(self, x):
        """Function returning the imaginary part of a result."""
        x_call_oop = x._call_out_of_place

        def imagpart_oop(x):
            return np.asarray(x_call_oop(x), dtype=self.out_dtype).imag

        if is_real_dtype(self.out_dtype):
            return self.zero()
        else:
            rdtype = real_dtype(self.out_dtype)
            rspace = self.astype(rdtype)
            return rspace.element(imagpart_oop)

    def _conj(self, x):
        """Function returning the complex conjugate of a result."""
        x_call_oop = x._call_out_of_place

        def conj_oop(x):
            return np.asarray(x_call_oop(x), dtype=self.out_dtype).conj()

        if is_real_dtype(self.out_dtype):
            return x
        else:
            return self.element(conj_oop)

    @property
    def examples(self):
        """Return example functions in the space.

        Example functions include:

        Zero
        One
        Heaviside function
        Hypercube characteristic function
        Hypersphere characteristic function
        Gaussian
        Linear gradients
        """
        # Get the points and calculate some statistics on them
        mins = self.domain.min()
        maxs = self.domain.max()
        means = (maxs + mins) / 2.0
        stds = (maxs - mins) / 4.0
        ndim = getattr(self.domain, 'ndim', None)

        # Zero and One
        yield ('Zero', self.zero())
        try:
            yield ('One', self.one())
        except NotImplementedError:
            pass

        # Indicator function in first dimension
        def _step_fun(x):
            if ndim == 1:
                return x > means[0]
            else:
                return (x[0] > means[0]) + 0 * sum(x[1:])  # fix size

        yield ('Step', self.element(_step_fun))

        # Indicator function on hypercube
        def _cube_fun(x):
            if ndim > 1:
                result = True
                for points, mean, std in zip(x, means, stds):
                    result = np.logical_and(result, points < mean + std)
                    result = np.logical_and(result, points > mean - std)
            else:
                result = np.logical_and(x < means + stds,
                                        x > means - stds)

            return result

        yield ('Cube', self.element(_cube_fun))

        # Indicator function on hypersphere
        if self.domain.ndim > 1:  # Only if ndim > 1, don't duplicate cube
            def _sphere_fun(x):
                if ndim == 1:
                    x = (x,)

                r = 0

                for points, mean, std in zip(x, means, stds):
                    r = r + (points - mean) ** 2 / std ** 2

                return r < 1.0

            yield ('Sphere', self.element(_sphere_fun))

        # Gaussian function
        def _gaussian_fun(x):
            if ndim == 1:
                x = (x,)

            r2 = 0
            for points, mean, std in zip(x, means, stds):
                r2 = r2 + (points - mean) ** 2 / ((std / 2) ** 2)

            return np.exp(-r2)

        yield ('Gaussian', self.element(_gaussian_fun))

        # Gradient in each dimensions
        for dim in range(self.domain.ndim):
            def _gradient_fun(x):
                if ndim == 1:
                    x = (x,)

                s = 0
                for ind in range(len(x)):
                    if ind == dim:
                        s = s + (x[ind] - mins[ind]) / (maxs[ind] - mins[ind])
                    else:
                        s = s + x[ind] * 0  # Correct broadcast size

                return s

            yield ('grad {}'.format(dim), self.element(_gradient_fun))

        # Gradient in all dimensions
        if self.domain.ndim > 1:  # Only if ndim > 1, don't duplicate grad 0
            def _all_gradient_fun(x):
                if ndim == 1:
                    x = (x,)

                s = 0
                for ind in range(len(x)):
                    s = s + (x[ind] - mins[ind]) / (maxs[ind] - mins[ind])

                return s

            yield ('Grad all', self.element(_all_gradient_fun))

    @property
    def element_type(self):
        """`FunctionSpaceElement`"""
        return FunctionSpaceElement

    def __repr__(self):
        """Return ``repr(self)``."""
        inner_str = '{!r}'.format(self.domain)
        dtype_str = dtype_repr(self.out_dtype)

        if self.field == RealNumbers():
            if self.out_dtype == np.dtype('float64'):
                pass
            else:
                inner_str += ', out_dtype={}'.format(dtype_str)

        elif self.field == ComplexNumbers():
            if self.out_dtype == np.dtype('complex128'):
                inner_str += ', field={!r}'.format(self.field)
            else:
                inner_str += ', out_dtype={}'.format(dtype_str)

        else:  # different field, name explicitly
            inner_str += ', field={!r}'.format(self.field)
            inner_str += ', out_dtype={}'.format(dtype_str)

        return '{}({})'.format(self.__class__.__name__, inner_str)

    def __str__(self):
        """Return ``str(self)``."""
        inner_str = '{}'.format(self.domain)
        dtype_str = dtype_repr(self.out_dtype)

        if self.field == RealNumbers():
            if self.out_dtype == np.dtype('float64'):
                pass
            else:
                inner_str += ', out_dtype={}'.format(dtype_str)

        elif self.field == ComplexNumbers():
            if self.out_dtype == np.dtype('complex128'):
                inner_str += ', field={!r}'.format(self.field)
            else:
                inner_str += ', out_dtype={}'.format(dtype_str)

        else:  # different field, name explicitly
            inner_str += ', field={!r}'.format(self.field)
            inner_str += ', out_dtype={}'.format(dtype_str)

        return '{}({})'.format(self.__class__.__name__, inner_str)


class FunctionSpaceElement(LinearSpaceElement, FunctionSetElement):

    """Representation of a `FunctionSpace` element."""

    def __init__(self, fspace, fcall):
        """Initialize a new instance.

        Parameters
        ----------
        fspace : `FunctionSpace`
            Set of functions this element lives in.
        fcall : callable
            The actual instruction for out-of-place evaluation.
            It must return an `FunctionSet.range` element or a
            ``numpy.ndarray`` of such (vectorized call).
        """
        if not isinstance(fspace, FunctionSpace):
            raise TypeError('`fspace` {!r} not a `FunctionSpace` '
                            'instance'.format(fspace))

        LinearSpaceElement.__init__(self, fspace)
        FunctionSetElement.__init__(self, fspace, fcall)

    # Tradeoff: either we subclass LinearSpaceElement first and override the
    # 3 methods in FunctionSetElement (as below) which LinearSpaceElement
    # also has, or we switch inheritance order and need to override all magic
    # methods from LinearSpaceElement which are not in-place. This is due to
    # the fact that FunctionSetElement inherits from Operator which defines
    # some of those magic methods, and those do not work in this case.
    __eq__ = FunctionSetElement.__eq__
    assign = FunctionSetElement.assign
    copy = FunctionSetElement.copy

    # Power functions are more general than the ones in LinearSpace
    def __pow__(self, p):
        """`f.__pow__(p) <==> f ** p`."""
        out = self.space.element()
        self.space._scalar_power(self, p, out=out)
        return out

    def __ipow__(self, p):
        """`f.__ipow__(p) <==> f **= p`."""
        return self.space._scalar_power(self, p, out=self)

    @property
    def real(self):
        """Pointwise real part of this function."""
        return self.space._realpart(self)

    @property
    def imag(self):
        """Pointwise imaginary part of this function."""
        return self.space._imagpart(self)

    def conj(self):
        """Pointwise complex conjugate of this function."""
        return self.space._conj(self)

    def __repr__(self):
        """Return ``repr(self)``."""
        return 'FunctionSpaceElement'

if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Simple iterative type optimization schemes."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()

from odl.operator import IdentityOperator, OperatorComp, OperatorSum


__all__ = ('landweber', 'conjugate_gradient', 'conjugate_gradient_normal',
           'gauss_newton')


# TODO: update all docs


def landweber(op, x, rhs, niter=1, omega=1, projection=None, callback=None):
    """Optimized implementation of Landweber's method.

    This method calculates an approximate least-squares solution of
    the inverse problem of the first kind

        :math:`\mathcal{A} (x) = y`,

    for a given :math:`y\\in \mathcal{Y}`, i.e. an approximate
    solution :math:`x^*` to

        :math:`\min_{x\\in \mathcal{X}}
        \\lVert \mathcal{A}(x) - y \\rVert_{\mathcal{Y}}^2`

    for a (Frechet-) differentiable operator
    :math:`\mathcal{A}: \mathcal{X} \\to \mathcal{Y}` between Hilbert
    spaces :math:`\mathcal{X}` and :math:`\mathcal{Y}`. The method
    starts from an initial guess :math:`x_0` and uses the
    iteration

    :math:`x_{k+1} = x_k -
    \omega \ \partial \mathcal{A}(x)^* (\mathcal{A}(x_k) - y)`,

    where :math:`\partial \mathcal{A}(x)` is the Frechet derivativ
    of :math:`\mathcal{A}` at :math:`x` and :math:`\omega` is a
    relaxation parameter. For linear problems, a choice
    :math:`0 < \omega < 2/\\lVert \mathcal{A}^2\\rVert` guarantees
    convergence, where :math:`\\lVert\mathcal{A}\\rVert` stands for the
    operator norm of :math:`\mathcal{A}`.

    Users may also optionally provide a projection to project each
    iterate onto some subset. For example enforcing positivity.

    This implementation uses a minimum amount of memory copies by
    applying re-usable temporaries and in-place evaluation.

    The method is also described in a
    `Wikipedia article
    <https://en.wikipedia.org/wiki/Landweber_iteration>`_.

    Parameters
    ----------
    op : `Operator`
        Operator in the inverse problem. It must have a `Operator.derivative`
        property, which returns a new operator which in turn has an
        `Operator.adjoint` property, i.e. ``op.derivative(x).adjoint`` must be
        well-defined for ``x`` in the operator domain.
    x : ``op.domain`` element
        Element to which the result is written. Its initial value is
        used as starting point of the iteration, and its values are
        updated in each iteration step.
    rhs : ``op.range`` element
        Right-hand side of the equation defining the inverse problem
    niter : int, optional
        Maximum number of iterations
    omega : positive float, optional
        Relaxation parameter in the iteration
    projection : callable, optional
        Function that can be used to modify the iterates in each iteration,
        for example enforcing positivity. The function should take one
        argument and modify it in-place.
    callback : callable, optional
        Object executing code per iteration, e.g. plotting each iterate

    Returns
    -------
    None
    """
    # TODO: add a book reference

    if x not in op.domain:
        raise TypeError('`x` {!r} is not in the domain of `op` {!r}'
                        ''.format(x, op.domain))

    # Reusable temporaries
    tmp_ran = op.range.element()
    tmp_dom = op.domain.element()

    for _ in range(niter):
        op(x, out=tmp_ran)
        tmp_ran -= rhs
        op.derivative(x).adjoint(tmp_ran, out=tmp_dom)
        x.lincomb(1, x, -omega, tmp_dom)

        if projection is not None:
            projection(x)

        if callback is not None:
            callback(x)


def conjugate_gradient(op, x, rhs, niter=1, callback=None):
    """Optimized implementation of CG for self-adjoint operators.

    This method solves the inverse problem (of the first kind)

    :math:`A x = y`

    for a linear and self-adjoint `Operator` ``A``.

    It uses a minimum amount of memory copies by applying re-usable
    temporaries and in-place evaluation.

    The method is described (for linear systems) in a
    `Wikipedia article
    <https://en.wikipedia.org/wiki/Conjugate_gradient_method>`_.

    Parameters
    ----------
    op : linear `Operator`
        Operator in the inverse problem. It must be linear and
        self-adjoint. This implies in particular that its domain and
        range are equal.
    x : ``op.domain`` element
        Element to which the result is written. Its initial value is
        used as starting point of the iteration, and its values are
        updated in each iteration step.
    rhs : ``op.range`` element
        Right-hand side of the equation defining the inverse problem
    niter : int, optional
        Maximum number of iterations
    callback : callable, optional
        Object executing code per iteration, e.g. plotting each iterate

    Returns
    -------
    None

    See Also
    --------
    conjugate_gradient_normal : Solver for nonsymmetric matrices
    """
    # TODO: add a book reference
    # TODO: update doc

    if op.domain != op.range:
        raise ValueError('operator needs to be self-adjoint')

    if x not in op.domain:
        raise TypeError('`x` {!r} is not in the domain of `op` {!r}'
                        ''.format(x, op.domain))

    r = op(x)
    r.lincomb(1, rhs, -1, r)       # r = rhs - A x
    p = r.copy()
    d = op.domain.element()  # Extra storage for storing A x

    sqnorm_r_old = r.norm() ** 2  # Only recalculate norm after update

    if sqnorm_r_old == 0:  # Return if no step forward
        return

    for _ in range(niter):
        op(p, out=d)  # d = A p

        inner_p_d = p.inner(d)

        if inner_p_d == 0.0:  # Return if step is 0
            return

        alpha = sqnorm_r_old / inner_p_d

        x.lincomb(1, x, alpha, p)            # x = x + alpha*p
        r.lincomb(1, r, -alpha, d)           # r = r - alpha*d

        sqnorm_r_new = r.norm() ** 2

        beta = sqnorm_r_new / sqnorm_r_old
        sqnorm_r_old = sqnorm_r_new

        p.lincomb(1, r, beta, p)                       # p = s + b * p

        if callback is not None:
            callback(x)


def conjugate_gradient_normal(op, x, rhs, niter=1, callback=None):
    """Optimized implementation of CG for the normal equation.

    This method solves the normal equation

    :math:`A^* A x = A^* y`

    to the inverse problem (of the first kind)

    :math:`A x = y`

    with a linear `Operator` ``A``.

    It uses a minimum amount of memory copies by applying re-usable
    temporaries and in-place evaluation.

    The method is described (for linear systems) in a
    `Wikipedia article
    <https://en.wikipedia.org/wiki/Conjugate_gradient_method#\
Conjugate_gradient_on_the_normal_equations>`_.

    Parameters
    ----------
    op : `Operator`
        Operator in the inverse problem. If not linear, it must have
        an implementation of `Operator.derivative`, which
        in turn must implement `Operator.adjoint`, i.e.
        the call ``op.derivative(x).adjoint`` must be valid.
    x : ``op.domain`` element
        Element to which the result is written. Its initial value is
        used as starting point of the iteration, and its values are
        updated in each iteration step.
    rhs : ``op.range`` element
        Right-hand side of the equation defining the inverse problem
    niter : int, optional
        Maximum number of iterations
    callback : callable, optional
        Object executing code per iteration, e.g. plotting each iterate

    Returns
    -------
    None

    See Also
    --------
    conjugate_gradient : Optimized solver for symmetric matrices
    """
    # TODO: add a book reference
    # TODO: update doc

    if x not in op.domain:
        raise TypeError('`x` {!r} is not in the domain of `op` {!r}'
                        ''.format(x, op.domain))

    d = op(x)
    d.lincomb(1, rhs, -1, d)               # d = rhs - A x
    p = op.derivative(x).adjoint(d)
    s = p.copy()
    q = op.range.element()
    sqnorm_s_old = s.norm() ** 2  # Only recalculate norm after update

    for _ in range(niter):
        op(p, out=q)                       # q = A p
        sqnorm_q = q.norm() ** 2
        if sqnorm_q == 0.0:  # Return if residual is 0
            return

        a = sqnorm_s_old / sqnorm_q
        x.lincomb(1, x, a, p)               # x = x + a*p
        d.lincomb(1, d, -a, q)              # d = d - a*Ap
        op.derivative(p).adjoint(d, out=s)  # s = A^T d

        sqnorm_s_new = s.norm() ** 2
        b = sqnorm_s_new / sqnorm_s_old
        sqnorm_s_old = sqnorm_s_new

        p.lincomb(1, s, b, p)               # p = s + b * p

        if callback is not None:
            callback(x)


def exp_zero_seq(base):
    """Default exponential zero sequence.

    It is defined by

        t_0 = 1.0
        t_m = t_(m-1) / base

    or, in closed form

        t_m = base^(-m-1)

    Parameters
    ----------
    base : float
        Base of the sequence. Its absolute value must be larger than
        1.

    Yields
    ------
    val : float
        The next value in the exponential sequence
    """
    value = 1.0
    while True:
        value /= base
        yield value


def gauss_newton(op, x, rhs, niter=1, zero_seq=exp_zero_seq(2.0),
                 callback=None):
    """Optimized implementation of a Gauss-Newton method.

    This method solves the inverse problem (of the first kind)

    :math:`A (x) = y`

    for a (Frechet-) differentiable `Operator` ``A`` using a
    Gauss-Newton iteration.

    It uses a minimum amount of memory copies by applying re-usable
    temporaries and in-place evaluation.

    A variant of the method applied to a specific problem is described
    in a
    `Wikipedia article
    <https://en.wikipedia.org/wiki/Gauss%E2%80%93Newton_algorithm>`_.

    Parameters
    ----------
    op : `Operator`
        Operator in the inverse problem. If not linear, it must have
        an implementation of `Operator.derivative`, which
        in turn must implement `Operator.adjoint`, i.e.
        the call ``op.derivative(x).adjoint`` must be valid.
    x : ``op.domain`` element
        Element to which the result is written. Its initial value is
        used as starting point of the iteration, and its values are
        updated in each iteration step.
    rhs : ``op.range`` element
        Right-hand side of the equation defining the inverse problem
    niter : int, optional
        Maximum number of iterations
    zero_seq : iterable, optional
        Zero sequence whose values are used for the regularization of
        the linearized problem in each Newton step
    callback : callable, optional
        Object executing code per iteration, e.g. plotting each iterate

    Returns
    -------
    None
    """
    if x not in op.domain:
        raise TypeError('`x` {!r} is not in the domain of `op` {!r}'
                        ''.format(x, op.domain))

    x0 = x.copy()
    id_op = IdentityOperator(op.domain)
    dx = op.domain.zero()

    tmp_dom = op.domain.element()
    u = op.domain.element()
    tmp_ran = op.range.element()
    v = op.range.element()

    for _ in range(niter):
        tm = next(zero_seq)
        deriv = op.derivative(x)
        deriv_adjoint = deriv.adjoint

        # v = rhs - op(x) - deriv(x0-x)
        # u = deriv.T(v)
        op(x, out=tmp_ran)              # eval  op(x)
        v.lincomb(1, rhs, -1, tmp_ran)  # assign  v = rhs - op(x)
        tmp_dom.lincomb(1, x0, -1, x)   # assign temp  tmp_dom = x0 - x
        deriv(tmp_dom, out=tmp_ran)     # eval  deriv(x0-x)
        v -= tmp_ran                    # assign  v = rhs-op(x)-deriv(x0-x)
        deriv_adjoint(v, out=u)         # eval/assign  u = deriv.T(v)

        # Solve equation Tikhonov regularized system
        # (deriv.T o deriv + tm * id_op)^-1 u = dx
        tikh_op = OperatorSum(OperatorComp(deriv.adjoint, deriv),
                              tm * id_op, tmp_dom)

        # TODO: allow user to select other method
        conjugate_gradient(tikh_op, dx, u, 3)

        # Update x
        x.lincomb(1, x0, 1, dx)  # x = x0 + dx

        if callback is not None:
            callback(x)

if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Douglas-Rachford splitting algorithm for convex optimization."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()

from odl.operator import Operator


__all__ = ('douglas_rachford_pd',)


def douglas_rachford_pd(x, f, g, L, tau, sigma, niter,
                        callback=None, **kwargs):
    """Douglas-Rachford primal-dual splitting algorithm.

    Minimizes the sum of several convex functions composed with linear
    operators::

        min_x f(x) + sum_i g_i(L_i x)

    where ``f``, ``g_i`` are convex functions, ``L_i`` are linear `Operator`'s.

    Can also be used to solve the more general problem::

        min_x f(x) + sum_i (g_i @ l_i)(L_i x)

    where ``l_i`` are convex functions and ``@`` is the infimal convolution::

        (g @ l)(x) = inf_y g(y) + l(x - y)

    Parameters
    ----------
    x : `LinearSpaceElement`
        Initial point, updated in-place.
    f : `Functional`
        `proximal factory` for the function ``f``.
    g : sequence of `Functional`'s
        Sequence of of the functions ``g_i``. Needs to have
        ``g[i].convex_conj.proximal``.
    L : sequence of `Operator`'s
        Sequence of `Opeartor`s` with as many elements as ``g``.
    tau : float
        Step size parameter for ``f``.
    sigma : sequence of floats
        Step size parameters for the ``g_i``s.
    niter : int
        Number of iterations.
    callback : callable, optional
        Function called with the current iterate after each iteration.

    Other Parameters
    ----------------
    l : sequence of `Functionals`'s, optional
        Sequence of of the functions ``l_i``. Needs to have
        ``l[i].convex_conj.proximal``.
        If omitted, the simpler problem without ``l_i``  will be considered.
    lam : float or callable, optional
        Overrelaxation step size. If callable, should take an index (zero
        indexed) and return the corresponding step size.

    Notes
    -----
    The mathematical problem to solve is

    .. math::
       \min_x f(x) + \sum_{i=0}^n (g_i \Box l_i)(L_i x),

    where :math:`f`, :math:`g_i`, :math:`l_i` are proper, convex and lower
    semicontinuous and :math:`L_i` are linear operators. The infimal
    convolution :math:`g \Box l` is defined by

    .. math::
       (g \Box l)(x) = \inf_y g(y) + l(x - y).

    The simplified problem,

    .. math::
        \min_x f(x) + \sum_{i=0}^n g_i(L_i x),

    can be obtained by setting

    .. math::
        l(x) = 0 \\text{ if } x = 0, \infty \\text{ else.}

    To guarantee convergence, the parameters :math:`\\tau`, :math:`\\sigma_i`
    and :math:`L_i` need to satisfy

    .. math::
       \\tau \\sum_{i=1}^n \\sigma_i ||L_i||^2 < 4

    The parameter :math:`\\lambda` needs to satisfy :math:`0 < \\lambda < 2`
    and if it is given as a function it needs to satisfy

    .. math::
        \\sum_{n=1}^\infty \\lambda_n (2 - \\lambda_n) = +\infty.

    See Also
    --------
    odl.solvers.nonsmooth.chambolle_pock.chambolle_pock_solver :
        Solver for similar problems.
    odl.solvers.nonsmooth.forward_backward.forward_backward_pd :
        Solver for similar problems which can additionaly handle a
        differentiable term.

    References
    ----------
    For references on the algorithm, see algorithm 3.1 in [BH2013]_.
    """

    # Problem size
    m = len(L)

    # Validate input
    if not all(isinstance(op, Operator) for op in L):
        raise ValueError('`L` not a sequence of operators')
    if not all(op.is_linear for op in L):
        raise ValueError('not all operators in `L` are linear')
    if not all(x in op.domain for op in L):
        raise ValueError('`x` not in the domain of all operators')
    if len(sigma) != m:
        raise ValueError('len(sigma) != len(L)')
    if len(g) != m:
        raise ValueError('len(prox_cc_g) != len(L)')

    prox_cc_g = [gi.convex_conj.proximal for gi in g]

    # Get parameters from kwargs
    l = kwargs.pop('l', None)
    if l is not None and len(l) != m:
        raise ValueError('`l` does not have the same number of '
                         'elements as `L`')
    if l is not None:
        prox_cc_l = [li.convex_conj.proximal for li in l]

    lam_in = kwargs.pop('lam', 1.0)
    if not callable(lam_in) and not (0 < lam_in < 2):
        raise ValueError('`lam` must callable or a number between 0 and 2')
    lam = lam_in if callable(lam_in) else lambda _: lam_in

    # Check for unused parameters
    if kwargs:
        raise TypeError('unexpected keyword argument: {}'.format(kwargs))

    # Pre-allocate values
    v = [Li.range.zero() for Li in L]
    p1 = x.space.zero()
    p2 = [Li.range.zero() for Li in L]
    z1 = x.space.zero()
    z2 = [Li.range.zero() for Li in L]
    w1 = x.space.zero()
    w2 = [Li.range.zero() for Li in L]

    for k in range(niter):
        tmp_1 = sum(Li.adjoint(vi) for Li, vi in zip(L, v))
        tmp_1.lincomb(1, x, -tau / 2, tmp_1)
        f.proximal(tau)(tmp_1, out=p1)
        w1.lincomb(2.0, p1, -1, x)

        for i in range(m):
            prox_cc_g[i](sigma[i])(v[i] + (sigma[i] / 2.0) * L[i](w1),
                                   out=p2[i])
            w2[i].lincomb(2.0, p2[i], -1, v[i])

        tmp_2 = sum(Li.adjoint(wi) for Li, wi in zip(L, w2))
        z1.lincomb(1.0, w1, - (tau / 2.0), tmp_2)
        x += lam(k) * (z1 - p1)

        for i in range(m):
            tmp = w2[i] + (sigma[i] / 2.0) * L[i](2.0 * z1 - w1)
            if l is not None:
                # In this case the infimal convolution is used.
                prox_cc_l[i](sigma[i])(tmp, out=z2[i])
            else:
                # If the infimal convolution is not given, prox_cc_l is the
                # identity and hence omitted. For more details, see the
                # documentation.
                z2[i].assign(tmp)
            v[i] += lam(k) * (z2[i] - p2[i])

        if callback is not None:
            callback(p1)

    # The final result is actually in p1 according to the algorithm, so we need
    # to assign here.
    x.assign(p1)
